from __future__ import annotations

import ast
from pathlib import Path
from typing import Literal, Self, cast

from pydantic import BaseModel, Field, model_validator

from ..tools.filesystem import Dir, FileSystem
from ..util import filter as lfilter, find, maybe_or_call, not_none
from .base import Module, ModuleDep

NameAndDoc = tuple[str, str | None]
CodeElements = tuple[list[NameAndDoc], list[NameAndDoc], list[str], str | None]


class PythonModule(Module):
    """A Python module."""

    src_lang: Literal['python'] = 'python'
    imports: list[AstImport | AstImportFrom]
    elements: CodeElements

    @classmethod
    def from_content(cls, relative_path: Path, content: str) -> Self:
        try:
            elements = _parse_code_elements(ast.parse(content))
        except Exception as e:
            print(f'Failed to parse code elements from {relative_path}: {e}')
            elements = ([], [], [], None)

        return cls(
            relative_path=relative_path,
            content=content,
            imports=parse_imports(content),
            elements=elements,
        )

    @model_validator(mode='after')
    def gen_name(self) -> Self:
        self.name = rel_path_to_python_module_name(self.relative_path)
        return self


def _parse_code_elements(tree: ast.Module) -> CodeElements:
    """Parse functions, classes, constants from a Python file

    Returns:
        functions (list[tuple[str, str | None]]): functions and their docstrings
        classes (list[tuple[str, str | None]]): classes and their docstrings
        constants (list[str]): constants
        docstring (str | None): file-level docstring
    """
    functions = []
    classes = []
    constants = []
    docstring = None

    # fmt: off
    def maybe_docstring(node):
        def maybe_head(x): return x[0] if len(x) > 0 else None
        def value_of_expr(x): return x.value if isinstance(x, ast.Expr) else None
        def value_of_const(x): return x.value if isinstance(x, ast.Constant) else None
        def str_val(x): return x if isinstance(x, str) else None
        return str_val(value_of_const(value_of_expr(maybe_head(node.body))))
    # fmt: on

    # Get module docstring
    docstring = maybe_docstring(tree)

    def name_and_docstring(node):
        return node.name, maybe_docstring(node)

    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef | ast.AsyncFunctionDef):
            node = cast(ast.FunctionDef | ast.AsyncFunctionDef, node)
            functions.append(name_and_docstring(node))
        elif isinstance(node, ast.ClassDef):
            classes.append(name_and_docstring(node))
        elif isinstance(node, ast.Assign):
            for target in node.targets:
                if isinstance(target, ast.Name) and target.id.isupper():
                    constants.append(target.id)

    return functions, classes, constants, docstring


def rel_path_to_python_module_name(path: Path) -> str:
    """
    Example:
    - `basic.py` -> `basic`
    - `utils/helpers.py` -> `utils.helpers`
    - `advanced/geometry/shapes.py` -> `advanced.geometry.shapes`
    """
    parts = list(path.parts)
    parts = [p.removesuffix('.py') for p in parts]
    return '.'.join(parts)


def build_py_dep_graph(fs: FileSystem) -> ModuleDep[PythonModule]:
    # Parse all Python modules
    py_modules: list[PythonModule] = []
    for node in fs.nodes:
        rel_path = node.relative_path
        if isinstance(node, Dir) or rel_path.suffix != '.py':
            continue
        py_modules.append(PythonModule.from_content(rel_path, node.content))

    adjacency_list: dict[str, set[str]] = {m.path_key: set() for m in py_modules}

    for py_module in py_modules:
        for import_ast in py_module.imports:
            targets = resolve_import(import_ast, py_module, py_modules, fs)

            for t in targets:
                if t != py_module:
                    adjacency_list[py_module.path_key].add(t.path_key)

    return ModuleDep.from_adjacency_list(py_modules, adjacency_list)


class AstImport(BaseModel):
    """Serializable ast.Import"""

    lineno: int = Field(description='Line number where import appears')
    col_offset: int = Field(description='Column offset where import appears')
    end_lineno: int | None = Field(
        default=None, description='Line number where import ends'
    )
    end_col_offset: int | None = Field(
        default=None, description='Column offset where import ends'
    )
    aliases: list[str] = Field(description='Aliases used in import')

    @classmethod
    def from_ast(cls, node: ast.Import) -> Self:
        return cls(
            lineno=node.lineno,
            col_offset=node.col_offset,
            end_lineno=node.end_lineno,
            end_col_offset=node.end_col_offset,
            aliases=[alias.name for alias in node.names],
        )


class AstImportFrom(AstImport):
    """Serializable ast.ImportFrom"""

    module: str | None = Field(description='Module being imported from')
    level: int = Field(description='Level of import (0=absolute, 1+=relative)')

    @classmethod
    def from_ast(cls, node: ast.ImportFrom) -> Self:
        return cls(
            lineno=node.lineno,
            col_offset=node.col_offset,
            end_lineno=node.end_lineno,
            end_col_offset=node.end_col_offset,
            aliases=[alias.name for alias in node.names],
            module=node.module,
            level=node.level,
        )


def parse_imports(content: str) -> list[AstImport | AstImportFrom]:
    """Parse import statements from a Python file"""
    imports = []
    tree = ast.parse(content)

    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            imports.append(AstImport.from_ast(node))

        elif isinstance(node, ast.ImportFrom):
            imports.append(AstImportFrom.from_ast(node))
    return imports


def resolve_import(
    import_ast: AstImport | AstImportFrom,
    importing_module: PythonModule,
    all_modules: list[PythonModule],
    fs: FileSystem,
) -> list[PythonModule]:
    """
    Resolve an import statement to 0+ target modules.
    """

    # fmt: off
    def map_filter(f, xs): return lfilter(not_none, map(f, xs))
    def path_key_is(p): return lambda m: m.path_key == p
    def find_module(p): return find(path_key_is(p), all_modules)
    # fmt: on

    def for_import_from(ast: AstImportFrom):
        # level 0: from pkg2 import b, c
        # level >0: from .foo import bar
        # fmt: off
        base = (ast.module if ast.level == 0 else
                resolve_relative_import_base_module(ast, importing_module))
        # fmt: on

        def import_from_alias(alias):
            # alias strings are names for ImportFrom (e.g. "b", or "*")
            name = alias.split('.')[0]

            # conservative: if star import, fall back to the base module/package itself
            # if resolved is None, fallback to depend on the base module/package
            # resolved_path is None or like "pkg2/b.py" or "pkg2/__init__.py"
            # fmt: off
            def find_base(): return find_module_by_name(cast(str, base), all_modules)
            return (find_base() if name == '*' else
                    maybe_or_call(find_base, find_module,
                                  resolve_from_import(fs, cast(str, base), name)))
            # fmt: on

        return map_filter(import_from_alias, import_ast.aliases) if base else []

    def for_plain_import():
        def for_import_alias(alias):
            # your AstImport stores alias.name strings
            # Prefer exact dotted match first (pkg2.b -> pkg2/b.py if present)
            dotted = alias.split('.')
            cand = str(Path(*dotted).with_suffix('.py'))
            # fmt: off
            mod_of_dotted = (None if fs.get_node_by_path(cand) is None else
                             find_module(cand))
            # fmt: on
            # or fall back to existing heuristics (root module / filename / etc)
            return mod_of_dotted or find_module_by_name(dotted[0], all_modules)

        return map_filter(for_import_alias, import_ast.aliases)

    # fmt: off
    return (for_import_from(import_ast) if isinstance(import_ast, AstImportFrom) else
            for_plain_import() if isinstance(import_ast, AstImport) else [])
    # fmt: on


def resolve_relative_import_base_module(
    import_ast: AstImportFrom,
    importing_module: PythonModule,
) -> str | None:
    """
    Convert a relative import (level>0) into an absolute-ish dotted base module name.

    Examples (importing_module = pkg1/a.py => pkg1.a):
      from .foo import bar   -> base module "pkg1.foo"
      from ..pkg2 import b   -> base module "pkg2"
      from . import x        -> base module "pkg1"
    """
    # importing module name: pkg1.a
    importing_name = cast(str, importing_module.name)
    importing_pkg_parts = importing_name.split('.')[:-1]  # package parts only

    # level=1 means "current package"
    up = max(import_ast.level - 1, 0)
    if up > len(importing_pkg_parts):
        return None

    base_parts = importing_pkg_parts[: len(importing_pkg_parts) - up]

    if import_ast.module:
        base_parts += import_ast.module.split('.')

    if not base_parts:
        return None

    return '.'.join(base_parts)


def find_module_by_name(
    target_name: str,
    all_modules: list[PythonModule],
) -> PythonModule | None:
    """Find a module by name, trying different matching strategies."""

    # Try exact match first
    for module in all_modules:
        if module.name == target_name:
            return module

    # Try matching just the filename (stem)
    for module in all_modules:
        if module.relative_path.stem == target_name:
            return module

    # Try matching as root of dotted name
    for module in all_modules:
        module_name = cast(str, module.name)
        if module_name.split('.')[0] == target_name:
            return module

    return None


def resolve_from_import(fs, base_mod: str, name: str) -> str | None:
    # base_mod like "pkg2", name like "b"
    cand_mod = Path(*base_mod.split('.')) / f'{name}.py'
    if fs.get_node_by_path(str(cand_mod)) is not None:
        return str(cand_mod)

    cand_pkg = Path(*base_mod.split('.')) / name / '__init__.py'
    if fs.get_node_by_path(str(cand_pkg)) is not None:
        return str(cand_pkg)

    # fall back: package itself
    base_pkg_init = Path(*base_mod.split('.')) / '__init__.py'
    if fs.get_node_by_path(str(base_pkg_init)) is not None:
        return str(base_pkg_init)

    base_file = Path(*base_mod.split('.')).with_suffix('.py')
    if fs.get_node_by_path(str(base_file)) is not None:
        return str(base_file)

    return None


def experiment():
    fs = FileSystem.from_disk(Path('data/code4'))

    py_dep = build_py_dep_graph(fs)

    for node in py_dep.nodes:
        print(node)
        print(f'imports: {node.imports}')
        print(f'relative_path: {node.relative_path}')
        print('')
        print(f'node.content: {node.content}')

    # print (py_dep.nodes)

    for e in py_dep.nx_graph.edges():
        print(e)


if __name__ == '__main__':
    # import pytest
    experiment()
    # pytest.main([__file__])



# ruff: noqa: E702
import json
import logging
import os
from collections.abc import Callable, Iterable
from itertools import groupby
from typing import TypeVar

T = TypeVar('T')
U = TypeVar('U')


# fmt: off
def member[T](xs: Iterable[T]) -> Callable[[T], bool]: return lambda x: x in xs
def fst[T, U](xy: tuple[T, U]) -> T: x, _ = xy; return x
def snd[T, U](xy: tuple[T, U]) -> U: _, y = xy; return y
def swap[T, U](xy: tuple[T, U]) -> tuple[U, T]: x, y = xy; return y, x
def dup[T](x: T) -> tuple[T, T]: return x, x
def distrib(xys): return [(x, y) for x, ys in xys for y in ys]
def not_none(x): return x is not None
def guard(f, x): return x if f(x) else None
def maybe[T, U](f: Callable[[T], U], x: T | None) -> U | None: return None if x is None else f(x)
def maybe_else[T, U](y: U, f: Callable[[T], U], x: T | None) -> U: return y if x is None else f(x)
def maybe_or_call[T, U](g: Callable[[], U], f: Callable[[T], U], x: T | None) -> U:
    return g() if x is None else f(x)
def head(x): return next(iter(x), None)
def filter(p, xs): return [x for x in xs if p(x)]
def find(p, xs): return next((x for x in xs if p(x)), None)
def rev(items): return list(reversed(list(items)))
def flat_map(f, xs): return [y for x in xs for y in f(x)]
def unzip(pairs): return [x for x, _ in pairs], [y for _, y in pairs]
def map[T, U](f: Callable[[T], U], xs: Iterable[T]) -> list[U]: return [f(x) for x in xs]
def maybe_min(items, **kwargs): return None if len(items) == 0 else min(items, **kwargs)
def dict_keys(d): return list(d.keys())
def dict_values(d): return list(d.values())
def translate(d, x): return d.get(x, x)
def map_values(f, d): return {k: f(v) for k, v in d.items()}
def consult_or(z, d): return lambda x: d.get(x, z)
def raise_(e): raise e
def splice[T](f: Callable[..., T]) -> Callable[[Iterable], T]: return lambda args: f(*args)
def json_read(f): return json.loads(text_read(f))
def text_read(f: str) -> str:
    with open(f) as s: return s.read()


def scanl(f: Callable[[U, T], U], s: U, xs: Iterable[T]) -> Iterable[U]:
    yield s
    for x in xs:
        s = f(s, x)
        yield s


def intersperse(sep, xs):
    yield next(xs)
    for x in xs:
        yield sep
        yield x

def group_snd_by_fst(pairs):
    return [(x, map(snd, y)) for x, y in groupby(sorted(pairs, key=fst), key=fst)]

def mdict(pairs): return dict(group_snd_by_fst(pairs))

def lazy(f):
    cell = [None]
    def g():
        if cell[0] is None:
            cell[0] = f()
        return cell[0]
    return g


def show(x):
    from rich.console import Console
    from rich.pretty import Pretty

    r = (x.__rich__() if hasattr(x, '__rich__') else
         x if hasattr(x, '__rich_console__') else
         Pretty(x))
    width = int(os.popen('stty size', 'r').read().split(' ')[1])
    with Console(width=width) as c:
        c.print(r)
# fmt: on


def get_imandra_uni_key() -> str:
    key = os.getenv('IMANDRA_UNI_KEY')
    if not key:
        raise ValueError("CodeLogician requires 'IMANDRA_UNI_KEY' to be set!")
    return key


def setup_logging(debug: bool = False):
    logging.basicConfig(
        level=logging.DEBUG if debug else logging.WARNING,
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
        handlers=[logging.FileHandler('cl.log'), logging.StreamHandler()],
    )



from pydantic import BaseModel, ConfigDict, Field


class StorageEntry(BaseModel):
    model_config = ConfigDict(ser_json_bytes='base64', val_json_bytes='base64')

    key: str = Field(description='the CA store key')
    value: bytes = Field(description='the stored value')


# We tend to generate this using `google.protobuf.json_format.MessageToDict`
# This converts protobuf bytes to base64 strings
class Art(BaseModel):
    model_config = ConfigDict(ser_json_bytes='base64', val_json_bytes='base64')

    kind: str = Field(description='The kind of artifact')
    data: bytes = Field(description='Serialized data, in twine')
    api_version: str = Field(
        description=(
            'Version of the API. This is mandatory and must match with the imandrax-api'
            ' library version.'
        )
    )
    storage: list[StorageEntry] = Field(
        default_factory=list, description='Additional definitions on the side'
    )



#
#   Imandra Inc.
#
#   convert_api_to_mkd.py
#

import json

with open('iml_api_reference_202510011126.json') as infile:
    contents = infile.read()

j = json.loads(contents)
data = {'base': []}

for record in j:
    if record['module'] == '':
        data['base'].append(record)
    else:
        if record['module'] not in data:
            data[record['module']] = []
        data[record['module']].append(record)

for idx, module_name in enumerate(data.keys()):
    name = module_name.capitalize()

    sStr = f"""----
title: {name} Module
description: {name} Module types and functions
order: {idx + 1}
----

"""
    for record in data[module_name]:
        signature = record['signature']
        name = record['name']
        type_ = record['type']
        doc = record['doc']
        pattern = record['pattern']

        if doc is None:
            doc = 'N\\A'
        if pattern is None:
            pattern = 'N\\A'

        sStr += f"""
## `{signature.replace('\n', ' ')}`
- name: {name}
- type: {type_}
- signature: `{signature}`
- doc: {doc}
- pattern: {pattern}

"""

    filepath = f'api/{module_name.lower()}.md'
    with open(filepath, 'w') as outfile:
        print(sStr, file=outfile)



#
#   Imandra Inc.
#
#   vgs.py
#

from rich.text import Text
from textual.containers import (
    HorizontalGroup,
    ScrollableContainer,
    VerticalGroup,
    VerticalScroll,
)
from textual.reactive import reactive
from textual.screen import Screen
from textual.widgets import Button, Footer, Label, Rule, Static

from codelogician.modeling.metamodel import MetaModel

from ..common import Border, MyHeader, bind


class HScroll(ScrollableContainer):
    DEFAULT_CSS = """
    HScroll {
        width: 150;
        height: auto;
        layout: vertical;
        overflow-y: hidden;
        overflow-x: scroll;
    }
    """


def named_vals(name_val_pairs):
    def hg(name, val):
        l = Static(f'[$primary]{name}[/]: ')
        v = Static(str(val))
        v.styles.width = '1fr'
        l.styles.width = 'auto'
        return HorizontalGroup(l, v)

    return VerticalGroup(*(hg(name, val) for name, val in name_val_pairs))


def vg_ui(vg):
    from rich.pretty import Pretty as RPretty

    def req_repr(req):
        return named_vals(
            [
                ('Src func names', req.src_func_names),
                ('IML func names', req.iml_func_names),
                ('Description', req.description),
                ('Logical statement', req.logical_statement),
            ]
        )

    def data_repr(data):
        return named_vals([('Predicate', data.predicate), ('Kind', data.kind)])

    def maybe_(f, x):
        return Static('None')
        # return maybe_else(Static("None"), f, x)

    def text_repr(x):
        return Text(repr(x))

    return VerticalGroup(
        Border('RawVerifyReq', maybe_(req_repr, vg.raw)),
        Border('VerifyReqData', maybe_(data_repr, vg.data)),
        Border(
            'Result',
            HScroll(Static(RPretty(vg.res, overflow='ellipsis', no_wrap=True))),
        ),
    )


class VGsScreen(Screen):
    """ """

    mmodel: reactive[None | MetaModel] = reactive(None, recompose=True)

    def on_mount(self):
        self.title = 'Verification Goals'
        bind(self, 'mmodel')

    # def watch_mmodel(self, old_value : MetaModel, new_value : MetaModel):
    #     pass

    def compose(self):
        """ """
        yield MyHeader()
        with VerticalScroll():
            if self.mmodel:
                for model_idx, path in enumerate(self.mmodel.models.keys()):
                    model = self.mmodel.models[path]
                    yield Rule()
                    yield Label(f'Model path: [$primary][b]{path}[/b][/]')
                    yield Static()
                    with HorizontalGroup():
                        yield Button('View model', id=f'view_{model_idx}')
                        yield Rule('vertical')
                        with VerticalScroll():
                            for i, vg in enumerate(model.verification_goals()):
                                yield Border(f'VG #{i + 1}', vg_ui(vg))

        yield Footer()



#
#   Imandra Inc.
#
#   codelogician/sketch/container.py
#

from pydantic import BaseModel
from rich.panel import Panel
from rich.pretty import Pretty

from .sketch import Sketch


class SketchContainer(BaseModel):
    """
    Container for helping us keep track of the sketches
    """

    sketches: dict[str, Sketch] = {}

    def add(self, new_sketch: Sketch) -> None:
        """
        Add spetch to the container
        """
        if new_sketch.sketch_id in self.sketches:
            raise ValueError(
                f'Sketch with specified ID is already present: {new_sketch.sketch_id}'
            )

        self.sketches[new_sketch.sketch_id] = new_sketch

    def from_id(self, sketch_id: str) -> Sketch | None:
        """
        Return sketch from sketch ID, None if it's missing
        """
        return self.sketches.get(sketch_id)

    def ids(self) -> list[str]:
        """
        List of available sketch IDs
        """
        return list(self.sketches.keys())

    def list_sketches(self) -> list[dict[str, str]]:
        """
        List available sketches
        """
        return [
            {
                'sketch_id': sketch_id,
                'anchor_model': self.sketches[sketch_id].anchor_model_path,
            }
            for sketch_id in self.sketches
        ]

    def toJSON(self):
        """
        Return JSON dict
        """
        return self.model_dump_json()

    @staticmethod
    def fromJSON(j: str | dict):
        """
        Create a SketchesContainer object from json dict
        """
        if isinstance(j, str):
            return SketchContainer.model_validate_json(j)
        else:
            return SketchContainer.model_validate(j)

    def __rich__(self):
        """Return the list of rich representations of the sketches we hold"""

        pretty = Pretty(self.toJSON(), indent_guides=True)
        return Panel(pretty, title='SketchContainer', border_style='green')



#
#   Imandra Inc.
#
#   state.py
#

import json
import logging
import os
import pathlib

from pydantic import BaseModel

from .config import ServerConfig

log = logging.getLogger(__name__)


class ServerState(BaseModel):
    """
    Overall server state - notice we just maintain list of directories for strategies.
    The actual states/etc. will be stored there.
    """

    abs_path: str
    strategy_paths: list[str] = []
    config: ServerConfig = ServerConfig()

    def save(self):
        """
        Save server state to disk
        """
        try:
            with open(self.abs_path, 'w') as outfile:
                json.dump(self.toJSON(), outfile)
        except Exception as e:
            raise Exception(f'Failed to write server state file to disk: {e}')

    def __str__(self):
        return str(self.toJSON())

    def __repr__(self):
        return str(self.toJSON())

    def toJSON(self):
        return self.model_dump_json()

    @staticmethod
    def fromDir(dir_path: str):
        """
        Load server state from config/command-line arguments
        """
        abs_path = pathlib.Path(dir_path).resolve()
        if not os.path.isdir(abs_path):
            raise Exception(f'Provided path [{abs_path}] is not a directory!')

        return ServerState.fromFile(os.path.join(abs_path, '.cl_server'))

    @staticmethod
    def fromFile(abs_path: str) -> 'ServerState':
        try:
            with open(abs_path) as infile:
                j = json.load(infile)
        except Exception as e:
            raise Exception(f'Failed to read in state file: [{abs_path}]; {e}')

        return ServerState.model_validate_json(j)


if __name__ == '__main__':
    pass



#
#   Imandra Inc.
#
#   codelogician/modeling/status.py
#

from enum import StrEnum

from imandra.u.agents.code_logician.base import FormalizationStatus
from pydantic import BaseModel

from .task import TaskKind


class SrcCodeStatus(StrEnum):
    SRC_CODE_CURRENT = 'Source_code_current'
    SRC_CODE_CHANGED = 'Source_code_changed'
    SRC_CODE_DELETED = 'Source_code_deleted'


class ModelStatus(BaseModel):
    """Represents the collective status of a model"""

    src_code_status: SrcCodeStatus
    instructions_added: bool
    deps_changed: bool
    deps_need_formalization: bool
    formalization_status: FormalizationStatus
    outstanding_task_ID: str | None = None
    outstanding_task_kind: TaskKind | None = None

    def src_code_changed(self):
        """Return True if source code changed"""
        return self.src_code_status == SrcCodeStatus.SRC_CODE_CHANGED

    def toJSON(self):
        """return a nice dict with the status"""
        return self.model_dump_json()

    @staticmethod
    def fromJSON(d):
        """Translated a JSON into ModelStatus object"""
        return ModelStatus.model_validate_json(d)

    def __repr__(self):
        s = 'Model Status:\n'
        s += f'Source code status: {self.src_code_status.name}\n'
        s += f'Instructions added: {"Yes" if self.instructions_added else "No"}\n'
        s += f'Dependencies changed: {"Yes" if self.deps_changed else "No"}\n'
        s += f'Formalization status: {self.formalization_status.name}\n'

        return s

    def __rich__(self):
        from rich.console import Group
        from rich.text import Text

        def yes_no(b):
            return 'Yes' if b else 'No'

        return Group(
            *(
                Text.assemble(label, (str(val), 'bold'))
                for label, val in [
                    ('Source code status: ', self.src_code_status),
                    ('Instructions added: ', yes_no(self.instructions_added)),
                    ('Dependencies changed: ', yes_no(self.deps_changed)),
                    ('Formalization task ID: ', str(self.outstanding_task_ID)),
                    ('Formalization task kind: ', str(self.outstanding_task_kind)),
                    ('Formalization status: ', self.formalization_status.__rich__()),
                ]
            )
        )



#
#   Imandra Inc.
#
#   docs.py
#

from itertools import islice
from pathlib import Path
from typing import Literal

import regex
from fuzzysearch import find_near_matches
from pydantic import BaseModel
from rich import print as printr
from rich.console import Group
from rich.markdown import Markdown
from rich.panel import Panel
from rich.text import Text
from rich.tree import Tree

from codelogician.util import filter, flat_map, fst, map, maybe, maybe_min, snd, splice


def abbrev(s):
    return s if len(s) <= 100 else s[:97] + '...'


class DocFile(BaseModel):
    """
    DocFile contains the information
    """

    title: str
    path: str
    description: str
    order: int
    content: str

    def topic_match(self, t: str) -> int | None:
        """
        Calculate distance between the specified topic and the title
        """

        res1 = find_near_matches(t.lower(), self.title.lower(), max_l_dist=3)
        res2 = find_near_matches(t.lower(), self.path.lower(), max_l_dist=2)
        return maybe_min([x.dist for x in res1 + res2])

    def __rich__(self):
        """
        Return a nice representation
        """

        return Panel(
            Group(
                Text(f'Name: {self.title}'),
                Text(f'Path: {Path(self.path).name}'),
                Text(f'Description: {self.description}'),
                Text('Content:'),
                Markdown(self.content),
            )
        )

    @staticmethod
    def get_lines_by_char_range(text: str, start_char_index: int, end_char_index: int):
        """
        Retrieves all lines from a given text that contain characters within
        the specified start and end character indexes.

        Args:
            text (str): The input text.
            start_char_index (int): The starting character index (inclusive).
            end_char_index (int): The ending character index (exclusive).

        Returns:
            list: A list of lines that contain characters within the specified range.
        """

        # Keep line endings for accurate indexing
        lines = text.splitlines(keepends=True)

        result_lines = []
        current_char_offset = 0

        for line in lines:
            line_start_char = current_char_offset
            line_end_char = current_char_offset + len(line)

            # Check for overlap between the requested range and the line's range
            if max(start_char_index, line_start_char) < min(
                end_char_index, line_end_char
            ):
                result_lines.append(line.strip('\n'))  # Remove newline for clean output

            current_char_offset += len(line)

        return '\n'.join(result_lines)

    def rich_selection(self, start_idx: int, end_idx: int, matched: str):
        """
        Print out the version of the
        """

        selection = self.get_lines_by_char_range(self.content, start_idx, end_idx)
        selection = selection.replace(matched, f'[bold]{matched}[/bold]')

        return Panel(
            Group(
                Text(f'Name: {self.title}'),
                Text(f'Path: {Path(self.path).name}'),
                Text(f'Matched: {matched}'),
                Text(f'Selected text:\n{selection}'),
            )
        )


def parse_markdown_or_code(path: str) -> DocFile:
    """
    Parse the markdown file and return a nice object
    """

    hline = regex.compile(r'\s*---+\s*')  # whitespace, at least 3 '-', whitespace
    lines = Path(path).read_text().split('\n')
    dividers = list(
        islice((i for i, l in enumerate(lines) if hline.fullmatch(l)), 0, 2)
    )

    if len(dividers) != 2:
        raise Exception(f'No frontmatter found in documentation {path}')

    first, second = dividers
    pairs = [l.split(':') for l in lines[first + 1 : second] if ':' in l]
    data = {k.strip().lower(): v for k, v in pairs}

    return DocFile(
        title=data['title'],
        order=int(data['order']),
        path=path,
        description=data['description'],
        content='\n'.join([*lines[:first], *lines[second + 1 :]])
        if len(lines) > second + 1
        else '',
    )


type Leaf = tuple[Literal['leaf'], tuple[str, DocFile]]
type Branch = tuple[Literal['branch'], tuple[tuple[str, DocFile], list[Node]]]
type Node = Leaf | Branch


def walk_directory(directory: Path) -> list[Node]:
    """
    Returns a tree, consisting of a list of nodes.
    Each `node` is either `('leaf', (str, DocFile))`, or
    or `('branch', ((str, DocFile), list[node]))`.
    """

    def maybe_parse(p: Path) -> DocFile | None:
        try:
            return parse_markdown_or_code(str(p))
        except Exception:
            return None

    # fmt: off
    def path_entry(p: Path):
        is_file = p.is_file()
        md = (maybe_parse(p / '_index.md') if not is_file else
              maybe_parse(p) if p.name != '_index.md' and p.suffix in ['.md', '.iml', '.py', '.ts']
              else None)
        # Sort order first then by is_file then filename
        return [] if md is None else [((md.order, is_file, p.name.lower()), (p, md))]

    def node(entry):
        (_, is_file, _), (path, md) = entry
        return (('leaf', (path, md)) if is_file else
                ('branch', ((path, md), sub_nodes(path))))

    def sub_nodes(path):
        return map(node, sorted(flat_map(path_entry, path.iterdir()), key=fst))

    return sub_nodes(Path(directory))


def leaves(nodes):
    def aux(kind, data):
        return [snd(data)] if kind == 'leaf' else leaves(snd(data))

    return flat_map(splice(aux), nodes)


def flatten_docs(nodes: list[Node]) -> str:
    def do_nodes(l, nodes):
        hd = '#' * l

        # fmt: off
        def fmt(_, doc): return f'{hd} {doc.title}\n{doc.content}'
        def do_leaf(data): return [fmt(*data)]
        def do_branch(data, nodes): return [fmt(*data), *do_nodes(l+1, nodes)]
        def do_node(kind, data):
            return (do_leaf if kind == 'leaf' else splice(do_branch))(data)
        # fmt: on

        return flat_map(splice(do_node), nodes)

    return '\n\n'.join(do_nodes(1, nodes))


def doc_tree(nodes) -> Tree:
    def dir_node(path, doc):
        label = f'[bold magenta]:open_file_folder: {abbrev(doc.description)}'
        return label, ('dim' if path.name.startswith('__') else '')

    def file_label(path, doc):
        text_filename = Text(path.name, 'green')
        text_filename.highlight_regex(r'\..*$', 'bold red')
        text_filename.stylize(f'link file://{path}')
        text_filename.append(f' {abbrev(doc.description)}', 'blue')
        suffix = path.suffix
        icon = '🐍 ' if suffix == '.py' else '📝 ' if suffix == '.iml' else '📄 '
        return Text(icon) + text_filename

    def add_nodes(nodes, tree):
        def add_leaf(data):
            tree.add(file_label(*data))

        def add_branch(data, nodes):
            text, style = dir_node(*data)
            add_nodes(nodes, tree.add(text, style=style, guide_style=style))

        for kind, data in nodes:
            {'leaf': add_leaf, 'branch': splice(add_branch)}[kind](data)

    tree = Tree('Documentation details', guide_style='bold bright_blue')
    add_nodes(nodes, tree)
    return tree


class DocManager:
    """
    Documentation Manager helper
    """

    def __init__(self, dir: str):
        nodes = walk_directory(Path(dir))
        self.doc_files = leaves(nodes)
        self.tree = doc_tree(nodes)

    def files(self) -> list[DocFile]:
        return self.doc_files

    # fmt: off
    def find_content(self, query: str, max_hits: int = 5) -> list[DocFile]:
        def doc_matches(d):
            return ([] if len(d.content) == 0 else
                    [(r, d) for r in find_near_matches(query, d.content, max_l_dist=1)])

        all_matches = flat_map(doc_matches, self.doc_files)
        return [doc.rich_selection(hit.start, hit.end, hit.matched)
                for hit, doc in sorted(all_matches, key=lambda x: fst(x).dist)[:max_hits]]

    def find_topic(self, topic: str, dist_cutoff: float = 5.0) -> DocFile | None:
        """
        Find DocFile for a specific topic (if possible)
        """
        topic = topic.replace('.md', '')

        def below_cutoff(dd):
            return fst(dd) is not None and fst(dd) <= dist_cutoff

        candidates = [(d.topic_match(topic), d) for d in self.doc_files]
        return maybe(snd, maybe_min(filter(below_cutoff, candidates), key=fst))
    # fmt: on

    def print_contents(self) -> None:
        """
        Print out the contents in a nice way...
        """

        printr(self.tree)



"""Advanced mathematical operations - Level 2 (depends on core.arithmetic)"""

from .arithmetic import add, divide, multiply, subtract
from .comparison import absolute_value


def square_root(x: float) -> float:
    """Calculate square root using Newton's method."""
    if x < 0:
        raise ValueError("Square root not defined for negative numbers")
    if x == 0:
        return 0.0

    guess = divide(x, 2.0)

    for _ in range(20):  # Simple iteration limit
        new_guess = divide(add(guess, divide(x, guess)), 2.0)
        if absolute_value(subtract(new_guess, guess)) < 1e-10:
            break
        guess = new_guess

    return new_guess


def factorial(n: int) -> int:
    """Calculate factorial of n."""
    if n < 0:
        raise ValueError("Factorial not defined for negative numbers")
    if n == 0 or n == 1:
        return 1

    result = 1
    for i in range(2, n + 1):
        result = int(multiply(result, i))
    return result


def is_even(n: int) -> bool:
    """Check if number is even."""
    return n % 2 == 0


def is_prime(n: int) -> bool:
    """Check if number is prime."""
    if n < 2:
        return False
    return all(n % i != 0 for i in range(2, int(square_root(n)) + 1))



#
#   Imandra Inc.
#
#   commands.py
#

from __future__ import annotations

from functools import partial
from pathlib import Path

from textual.app import App, ComposeResult
from textual.command import Hit, Hits, Provider
from textual.containers import VerticalScroll
from textual.widgets import Static

# Doesn't type check, unused
# class ViewCommandsProvider(Provider):
#     """A command provider for viewing model data"""

#     async def search(self, query: str) -> Hits:
#         """Search for appropriate server commands"""
#         _ = self.matcher(query)


class ServerCommandsProvider(Provider):
    """A command provider for sending commands to the server"""

    async def search(self, query: str) -> Hits:
        """Search for appropriate server commands"""

        matcher = self.matcher(query)
        for cmd in ['autoformalize']:
            command = f'cmd cl {cmd}'
            score = matcher.match(command)
            if score > 0:
                yield Hit(
                    score,
                    matcher.highlight(command),
                    partial(self.app.execute_server_command, command),  # pyright: ignore
                    help='Autoformalize is doing this!',
                )


class ModelCommandsProvider(Provider):
    """
    A command provider to open a Python file in the current working directory.
    """

    async def search(self, query: str) -> Hits:
        """Search for Python files."""
        matcher = self.matcher(query)

        model_paths = [
            'one.py',
            'two.py',
            'three.py',
            'one/one.py',
            'one/seven.py',
            'two/three/four/five.py',
        ]

        app = self.app
        assert isinstance(app, ViewerApp)

        cmds = ['autoformalize', 'set_context', 'something else']

        for path in model_paths:
            for cmd in cmds:
                command = f'cmd model {cmd} {path}'
                score = matcher.match(command)
                if score > 0:
                    yield Hit(
                        score,
                        matcher.highlight(command),
                        partial(self.app.execute_model_command, cmd, path),  # pyright: ignore
                        help='Help about model commands!',
                    )


class ViewerApp(App):
    """
    Demonstrate a command source.
    """

    COMMANDS = App.COMMANDS | {ServerCommandsProvider} | {ModelCommandsProvider}

    def compose(self) -> ComposeResult:
        with VerticalScroll():
            yield Static(id='code', expand=True)

    def execute_server_command(self, cmd: str) -> None:
        """We should execute server command here"""
        pass

    def execute_model_command(self, cmd: str, path: Path) -> None:
        """We should execute a command model."""
        pass

    def open_file(self, path: Path) -> None:
        """Open and display a file with syntax highlighting."""
        from rich.syntax import Syntax

        syntax = Syntax.from_path(
            str(path),
            line_numbers=True,
            word_wrap=False,
            indent_guides=True,
            theme='github-dark',
        )
        self.query_one('#code', Static).update(syntax)



#
#   Imandra Inc.
#
#   codelogician/sketch/sketch.py
#

import uuid
from pathlib import Path

from imandra.u.agents.code_logician.base import FormalizationStatus
from imandra.u.agents.code_logician.base.region_decomp import RegionDecomp
from imandra.u.agents.code_logician.base.vg import VG
from pydantic import BaseModel, Field

from codelogician.util import find, maybe, maybe_else

from ..modeling.model import Model
from ..tools.iml_utils import (
    add_definition,
    extract_definitions,
    remove_definition,
    topo_sort_definitions,
    update_definition,
)
from .task import (
    SketchChange,
    SketchChangeResult,
    SketchChangeTask,
    SketchChgDeleteDef,
    SketchChgInsertDef,
    SketchChgModifyDef,
    SketchChgSetModel,
)


def _iml_path(rel_path: str) -> str:
    return str(Path(rel_path).with_suffix('.iml'))


def _norm_def(body: str) -> str:
    return body.strip() + '\n'


def _ensure_trailing_nl(s: str) -> str:
    s2 = s.rstrip()
    return (s2 + '\n') if s2 else ''


class VGChange(BaseModel):
    """
    Helps us track VG changes over different states
    """

    is_new: bool  # Is this a new VG that was just added?
    is_removed: bool  # Has this VG been removed?
    res_changed: tuple[str, str] | None = None  # old vs new result type
    vg: VG  # Actual VG info


class ChangeSummary(BaseModel):
    """
    Summary of the formalization state changes since the last state where we had an admissable model
    """

    vg_changes: list[VGChange]

    def stats(self) -> dict[str, int]:
        """
        Return VG change statistics
        """

        def status_to_int(s: str) -> int:
            if s == 'unknown':
                return 1
            elif s == 'error' or s == 'err':
                return 2
            elif s == 'refuted':
                return 3
            elif s == 'verified_upto':
                return 4
            elif s == 'proved':
                return 5
            else:
                raise Exception(f'Unknown VG status: >{s}<')

        improved, deteriorated, proved, refuted = 0, 0, 0, 0
        for vg_change in self.vg_changes:
            if vg_change.is_removed or vg_change.res_changed is None:
                continue

            if vg_change.res_changed[1] in ['proved', 'verified_upto']:
                proved += 1
            if vg_change.res_changed[1] in ['refuted']:
                refuted += 1

            if not vg_change.is_new:
                if status_to_int(vg_change.res_changed[0]) > status_to_int(
                    vg_change.res_changed[1]
                ):
                    deteriorated += 1
                elif status_to_int(vg_change.res_changed[0]) > status_to_int(
                    vg_change.res_changed[1]
                ):
                    improved += 1

        return {
            'added': sum(1 if vgc.is_new else 0 for vgc in self.vg_changes),
            'removed': sum(1 if vgc.is_removed else 0 for vgc in self.vg_changes),
            'improved': improved,
            'deteriorated': deteriorated,
            'proved': proved,
            'refuted': refuted,
        }


def calc_vg_difference(old_vgs: list[VG], new_vgs: list[VG]) -> list[VGChange]:
    """
    Compute difference between two lists of VGs
    """
    vg_diff: list[VGChange] = []

    def find_a_match(new_v: VG, other_vgs: list[VG]) -> VG | None:
        def aux(new_data):
            def eq_preds(data):
                return data.predicate == new_data.predicate

            return find(lambda x: maybe_else(False, eq_preds, x.data), other_vgs)

        return maybe(aux, new_v.data)

    for new_vg in new_vgs:
        if new_vg.res is None:
            continue

        existing_vg = find_a_match(new_vg, old_vgs)

        if existing_vg is None:
            vg_diff.append(
                VGChange(
                    is_new=True,
                    is_removed=False,
                    res_changed=('', new_vg.res.res_type),
                    vg=new_vg,
                )
            )
        else:
            # This is an existing VG, let's see if its result type has changed at all...
            if (
                existing_vg.res is not None
                and new_vg.res.res_type != existing_vg.res.res_type
            ):
                vg_diff.append(
                    VGChange(
                        is_new=False,
                        is_removed=False,
                        res_changed=(existing_vg.res.res_type, new_vg.res.res_type),
                        vg=new_vg,
                    )
                )
            # ok, nothing changed
            else:
                vg_diff.append(
                    VGChange(
                        is_new=False, is_removed=False, res_changed=None, vg=new_vg
                    )
                )

    for old_vg in old_vgs:
        if old_vg.res is None:
            continue

        if find_a_match(old_vg, new_vgs) is None:
            vg_diff.append(
                VGChange(
                    is_new=False,
                    is_removed=True,
                    res_changed=(old_vg.res.res_type, ''),
                    vg=old_vg,
                )
            )

    return vg_diff


class SketchState(BaseModel):
    """
    SketchState contains the current version of
    """

    state_id: int  # State ID
    change: SketchChange  # Change that caused this state
    iml_code: str  # Actual IML code
    status: FormalizationStatus  # Formalization status
    error: str | None = None  # Error feedback from ImandraX

    vgs: list[VG]  # List of current verification goals
    old_vgs: list[VG]  # List of VGs in the last (good) state with admissable model
    decomps: list[RegionDecomp] = []  # Dictionary of decomposition requests/results

    vg_changes: list[VGChange] = Field(
        default_factory=lambda data: calc_vg_difference(
            old_vgs=data['old_vgs'],
            new_vgs=data['vgs'],
        )
        if 'vgs' in data and 'old_vgs' in data
        else []
    )

    # Summary of VG changes
    summary: ChangeSummary = Field(
        default_factory=lambda data: ChangeSummary(vg_changes=data['vg_changes'])
    )


class Sketch(BaseModel):
    """
    Sketch represent versions of a model
    """

    sketch_id: str = Field(default_factory=lambda: str(uuid.uuid4())[:8])
    processing: bool = False  # TODO do we need this for the UI?
    anchor_model_path: str  # relative path of the model that was used as the anchor
    init_iml_model: str = ''  # the initial model
    sketch_states: list[SketchState] = []  # here we keep track of changes in the state
    base_models: dict[
        str, Model
    ]  # Dict (by relative path) of base models that went into this
    def_owner: dict[str, str] = Field(
        default_factory=dict
    )  # def_name -> rel_path (e.g. "a.py")

    def get_state_by_idx(self, idx: int) -> SketchState:
        """
        Return SketchState for specific state ID
        """

        if not (0 <= idx and idx < len(self.sketch_states)):
            raise ValueError(f'Out of bounds state id: {idx}')

        return self.sketch_states[idx]

    def rollback(self, target_state_id: int | None = None):
        """
        rollback states
        """

        if target_state_id is None:
            # We should just be able to
            if len(self.sketch_states) in [0, 1]:
                raise Exception(
                    f'Not enough states[={len(self.sketch_states)}] to rollback!'
                )
            self.sketch_states = self.sketch_states[1:]

        else:
            if not (0 <= target_state_id and target_state_id < len(self.sketch_states)):
                raise Exception(
                    f'Invalid target state id[={target_state_id}] specified!'
                )

            self.sketch_states = self.sketch_states[: (target_state_id + 1)]

    def state_ids(self) -> list[int]:
        """
        Return list of SketchState IDs
        """
        if self.sketch_states:
            return list(range(len(self.sketch_states)))
        return []

    def latest_state(self) -> SketchState | None:
        """
        Return the latest sketch state if it's available
        """
        return self.sketch_states[0] if len(self.sketch_states) else None

    def last_good_state(self) -> SketchState | None:
        """
        Return last sketch state that has an admissable model
        """
        for i in range(len(self.sketch_states) - 1, -1, -1):
            if self.sketch_states[i].status == FormalizationStatus.TRANSPARENT:
                return self.sketch_states[i]

        return None

    def gen_init_task(self) -> SketchChangeTask:
        """
        This is the task that needs to be run to populate the initial state
        """
        return SketchChangeTask(
            sketch_id=self.sketch_id,
            iml_code=self.init_iml_model,
            change=SketchChgSetModel(new_iml_code=self.init_iml_model),
        )

    def get_task(self, change: SketchChange) -> SketchChangeTask:
        """
        Return a SketchChangeTask that can be sent to the worker to be analyzed by ImandraX
        """

        # Writing it this way so it's easier for pyright to pick up
        state = self.latest_state()

        if state is None:
            raise Exception('Attempting to create a task but we have no initial state!')

        new_iml_model = self.process_change(state.iml_code, change)

        # Now we have a sketch change task!
        return SketchChangeTask(
            sketch_id=self.sketch_id, change=change, iml_code=new_iml_model
        )

    def apply_change_result(self, change_result: SketchChangeResult) -> ChangeSummary:
        """
        Apply the result of running change task to the state and get the ChangeSummary as the result
        """
        if len(self.sketch_states) == 0 or self.last_good_state() is None:
            state_id = 0
            old_vgs = []
            self.init_iml_model = change_result.task.iml_code
        else:
            state_id = len(self.sketch_states) + 1
            old_vgs = self.last_good_state().vgs  # pyright: ignore

        new_state = SketchState(
            state_id=state_id,
            change=change_result.task.change,
            iml_code=change_result.task.iml_code,
            status=FormalizationStatus.TRANSPARENT
            if change_result.success
            else FormalizationStatus.INADMISSIBLE,
            error=change_result.error,
            vgs=change_result.vgs,
            old_vgs=old_vgs,
            decomps=change_result.decomps,
        )

        self.sketch_states.insert(0, new_state)

        return new_state.summary

    def _render_defs(self, defs: dict[str, str], order: list[str]) -> str:
        if not order:
            return ''
        # bodies already tend to end with "\n" from extract_definitions
        text = ''.join(defs[n] for n in order).rstrip() + '\n'
        return text

    def get_model_changes(self, *, misc_file_name: str = 'misc.iml') -> dict[str, str]:
        st = self.last_good_state()
        if st is None:
            return {}

        combined = st.iml_code.strip()
        combined_defs = extract_definitions(combined)
        combined_defs = {name: _norm_def(body) for name, body in combined_defs.items()}

        # Originals per-file, extracted from base models
        originals: dict[str, str] = {}
        orig_defs_by_file: dict[str, dict[str, str]] = {}

        for rel_path, m in self.base_models.items():
            p = _iml_path(rel_path)
            text = m.iml_model() or m.iml_code() or ''
            originals[p] = _ensure_trailing_nl(text)
            orig_defs_by_file[p] = extract_definitions(text)

        # Track per-file updated defs
        updated_defs_by_file = {
            p: dict(orig_defs_by_file[p]) for p in orig_defs_by_file
        }

        # Apply modifications back to owner file
        for name, new_body in combined_defs.items():
            owner_rel = self.def_owner.get(name)
            if owner_rel is None:
                # new definition: goes to misc
                continue

            owner_file = _iml_path(owner_rel)
            old_body = orig_defs_by_file.get(owner_file, {}).get(name)

            if old_body is None:
                # definition existed in combined but not found in owner file; treat as new -> misc
                continue

            if _norm_def(old_body) != new_body:
                updated_defs_by_file[owner_file][name] = new_body

        # Handle deletions: if a def existed in owner file but is not in combined, remove it
        for name, owner_rel in self.def_owner.items():
            owner_file = _iml_path(owner_rel)
            if (
                name in orig_defs_by_file.get(owner_file, {})
                and name not in combined_defs
            ):
                updated_defs_by_file[owner_file].pop(name, None)

        # New defs -> misc.iml
        new_defs = {n: b for n, b in combined_defs.items() if n not in self.def_owner}
        if new_defs:
            misc_text = originals.get(misc_file_name, '')
            misc_defs = extract_definitions(misc_text)
            misc_defs.update(new_defs)
            sorted_misc = topo_sort_definitions(misc_defs) if misc_defs else []
            originals[misc_file_name] = _ensure_trailing_nl(
                ''.join(misc_defs[n] for n in sorted_misc)
            )

        # Re-render each file in topo order and emit only diffs
        out: dict[str, str] = {}
        for file_path, defs in updated_defs_by_file.items():
            sorted_names = topo_sort_definitions(defs) if defs else []
            new_text = self._render_defs(defs, sorted_names)
            old_text = originals.get(file_path, '')
            if new_text != old_text:
                out[file_path] = new_text

        # If misc got new defs and differs
        if misc_file_name in originals and misc_file_name not in out:
            # if we created/updated misc above, compare to old (need old snapshot; simplest: only write if new_defs)
            if new_defs:
                out[misc_file_name] = originals[misc_file_name]

        return out

    def process_change(self, iml_code: str, change: SketchChange) -> str:
        """
        Process the change command for the specified code and return its updated version
        """

        if isinstance(change, SketchChgSetModel):
            updated_iml_code = change.new_iml_code

        elif isinstance(change, SketchChgInsertDef):
            try:
                updated_iml_code = add_definition(
                    source=iml_code, new_def=change.new_def_code
                )
            except Exception as e:
                raise Exception(f'Failed to insert: {e}')

        elif isinstance(change, SketchChgModifyDef):
            try:
                if isinstance(change.def_name_or_vg_idx, int):
                    updated_iml_code = update_definition(
                        source=iml_code,
                        name='',
                        new_def=change.new_def_body,
                        verify_index=change.def_name_or_vg_idx,
                    )
                else:
                    updated_iml_code = update_definition(
                        source=iml_code,
                        name=change.def_name_or_vg_idx,
                        new_def=change.new_def_body,
                    )
            except Exception as e:
                raise Exception(f'Failed to update definition: {e}')

        elif isinstance(change, SketchChgDeleteDef):
            try:
                if isinstance(change.def_name_or_vg_idx, int):
                    updated_iml_code = remove_definition(
                        source=iml_code, name='', verify_index=change.def_name_or_vg_idx
                    )
                else:
                    updated_iml_code = remove_definition(
                        source=iml_code, name=change.def_name_or_vg_idx
                    )
            except Exception as e:
                raise Exception(f'Failed to delete definition: {e}')

        else:
            raise Exception(f'Unsupported change event: {change}')

        return updated_iml_code

    def _baseline_iml_by_file(self) -> dict[str, str]:
        """
        Build the 'originals' dict for sync_definitions_topo.

        Keys here should be stable file identifiers. I recommend using `.iml` paths
        derived from each base model's rel_path.
        """
        originals: dict[str, str] = {}
        for rel_path, m in self.base_models.items():
            # Prefer "model-only" IML if you mean to patch defs, not artifacts.
            # If you only have iml_code(), that's ok too.
            src = m.iml_model() or m.iml_code() or ''
            iml_path = str(Path(rel_path).with_suffix('.iml'))
            originals[iml_path] = src.strip()
        return originals

    def __repr__(self):
        return f'Sketch [ID = {self.sketch_id}; Anchor model path = {self.anchor_model_path}]'

    @staticmethod
    def build_from_models(models: list[Model], *, anchor_model_path: str) -> 'Sketch':
        base_models = {m.rel_path: m for m in models}

        combined_defs: dict[str, str] = {}
        def_owner: dict[str, str] = {}
        dup_owners: dict[str, set[str]] = {}

        for m in models:
            src = (m.iml_model() or m.iml_code() or '').strip()
            if not src:
                continue

            defs = extract_definitions(src)
            for name, body in defs.items():
                # Detect ambiguous ownership (same top-level name defined in >1 base model)
                existing_owner = def_owner.get(name)
                if existing_owner is not None and existing_owner != m.rel_path:
                    dup_owners.setdefault(name, set()).update(
                        {existing_owner, m.rel_path}
                    )
                    continue

                # Keep first body we saw (content doesn't matter once ownership is unique)
                if name not in combined_defs:
                    combined_defs[name] = body
                    def_owner[name] = m.rel_path

        if dup_owners:
            parts: list[str] = []
            for name in sorted(dup_owners.keys()):
                owners = ', '.join(sorted(dup_owners[name]))
                parts.append(f'{name}: {owners}')
            raise ValueError(
                'Duplicate top-level definition names across base models: '
                + '; '.join(parts)
            )

        sorted_names = topo_sort_definitions(combined_defs)
        init_iml_model = ''.join(combined_defs[n] for n in sorted_names).strip() + '\n'

        sk = Sketch(
            anchor_model_path=anchor_model_path,
            init_iml_model=init_iml_model,
            base_models=base_models,
        )
        sk.def_owner = def_owner
        return sk



#
#   Imandra Inc.
#
#   config.py
#

import json
import logging

import yaml
from pydantic import BaseModel
from rich.text import Text

from ..strategy.config import StratConfig

log = logging.getLogger(__name__)


class ServerConfig(BaseModel):
    """Container for the Server configuration"""

    debug: bool = True  # Are we running in debug mode?
    host: str = '127.0.0.1'  # Host address
    port: int = 8000  # Port address
    mcp: bool = True  # Are we running the MCP server as well?

    strategy_configs: dict[str, StratConfig] = {}

    def strat_config(self, strat_name: str):
        """
        Return strategy configuration for a specified strategy
        """
        if strat_name in self.strategy_configs:
            return self.strategy_configs[strat_name]
        else:
            return StratConfig()

    # raise ValueError(f"{strat_name} not found!")

    def __rich__(self):
        """Create a Rich representation"""
        return Text(str(self))

    def toJSON(self):
        """ """
        return self.model_dump_json()

    def __repr__(self):
        """Return a nice representation of the Config object"""
        return json.dumps(self.toJSON(), indent=4)

    @staticmethod
    def fromYAML(path: str) -> 'ServerConfig':
        """Read-in the YAML file"""

        try:
            with open(path) as infile:
                config = yaml.safe_load(infile)
        except FileNotFoundError as e:
            log.warning(f'Failed to read in the YAML configuration: {str(e)}')
            raise Exception(f'No file found: {path}')
        except yaml.YAMLError as e:
            log.warning(f'Failed to parse the YAML file: {str(e)}')
            raise Exception(f'Failed to parse the YAML file: {str(e)}')
        except Exception as e:
            log.warning(f'Error parsing config: {str(e)}')
            raise Exception(f'Error parsing config: {str(e)}')

        # We should have a nice config container now
        try:
            config = ServerConfig.model_validate(config)
        except Exception as e:
            errMsg = f'ServerConfig validation error: {str(e)}'
            log.warning(errMsg)
            raise Exception(errMsg)

        return config



#
#   Imandra Inc.
#
#   events.py
#

import datetime

from imandra.u.agents.code_logician.command import Command
from pydantic import BaseModel, Field

from ..modeling.task import ModelTask
from ..sketch.task import SketchChange, SketchChangeResult
from .cl_agent_state import CLResult
from .config import StratConfigUpdate, StratMode


class StrategyEvent(BaseModel):
    """
    Base class
    """

    time: datetime.datetime = Field(default_factory=datetime.datetime.now)


class RunOneshotEvent(StrategyEvent, BaseModel):
    """
    Make strategy run in one-shot mode
    """

    pass  # We dont really need anything here...


class StratParamUpdate(StrategyEvent, BaseModel):
    """
    Event that updates strategy parameters
    """

    param: StratConfigUpdate


class ChangeMode(StrategyEvent, BaseModel):
    """
    Changes the mode (Idle/Auto)
    """

    new_mode: StratMode


class CLResultEvent(StrategyEvent, BaseModel):
    """
    Result that is generated by the CL worker when it's done processing the result
    """

    result: CLResult


class ModelCLTaskEvent(StrategyEvent, BaseModel):
    """
    Request to perform a specific command for all the workers
    """

    rel_path: str
    cmd: Command


class AutoModeCLTaskEvent(StrategyEvent, BaseModel):
    """
    Model formalization task that's generated by the strategy in AUTO mode
    """

    task: ModelTask


class SketchChangeEvent(StrategyEvent, BaseModel):
    """
    Change event for Sketch
    """

    sketch_id: str
    change: SketchChange


class SketchChangeResultEvent(StrategyEvent, BaseModel):
    """
    Result of running ImandraX on the latest change to a sketch
    """

    sketch_id: str
    change_result: SketchChangeResult



#
#   Imandra Inc.
#
#   step_view.py
#

from rich.text import Text
from textual.containers import VerticalGroup
from textual.widgets import (
    Collapsible,
    DataTable,
    Label,
    Pretty,
    Static,
    TabbedContent,
    TabPane,
)


def command_widget(command):
    args = command.model_dump(exclude={'type'})

    # Arguments table (if any)
    def do_args(args):
        max_width = 40
        yield Label('[b]Parameters:[/b]')

        # fmt: off
        def truncate(s): return s[:max_width - 3] + '...' if len(s) > max_width - 3 else s
        def fmt_list(l): return f'[{len(value)} items]' if l else '[]'
        def fmt_dict(l): return f'{{{len(value)} items}}' if l else '{}'
        def fmt_bool(t, f): return lambda b: f'[{t}]{b}[/{t}]' if b else f'[{f}]{b}[/{f}]'

        args_table = DataTable(show_header=False)
        for key, value in args.items():
            fmt = (truncate if isinstance(value, str) else
                   fmt_list if isinstance(value, list) else
                   fmt_dict if isinstance(value, dict) else
                   fmt_bool('bright_green','bright_red') if isinstance(value, bool) else
                   str)
            args_table.add_column('name')
            args_table.add_column('value')
            args_table.add_row(f'[dim]{key}:[/dim]', fmt(value))
        yield args_table
        # fmt: on

    cmd_name = command.type
    return VerticalGroup(
        Label(f'[b]Command: {cmd_name}[/b]'),
        *(do_args(args) if args else [Label('\n[dim]No parameters[/dim]')]),
    )


class TaskView(VerticalGroup):
    DEFAULT_CSS = """
        TaskView {
        border: round $foreground-muted;
        border-title-align: center;
        padding: 0 1 0 1;
        }"""

    def __init__(self, task):
        def trajectory_summary(formalizations):
            table = DataTable()  # title="Summary of Formalizations")
            table.add_column('#', width=2)
            table.add_column('Action', width=25)
            table.add_column('Status', width=15)
            table.add_column('Time', width=20)
            for i, f in enumerate(formalizations, 1):
                table.add_row(
                    str(i),
                    f.action,
                    f.fstate.status.__rich__(),
                    f.time_str,
                )
            yield table
            for f in formalizations:
                yield Static(f.__rich__())

        def formalizations_view(formalizations):
            with VerticalGroup():
                for i, f in enumerate(formalizations, 1):
                    title = (
                        Text(f'{i} {f.action} ')
                        + f.fstate.status.__rich__()
                        + Text(' ' + f.time_str)
                    )
                    yield Collapsible(Static(f.__rich__()), title=title)

        def summary_parts():
            yield Static(Text('Status: ', 'bold') + task.render_status())

            if task.precheck_failures:
                yield Static(
                    Text(
                        f'\nPrecheck Failures ({len(task.precheck_failures)}):',
                        style='bold bright_red',
                    )
                )
                for failure in task.precheck_failures:
                    yield Static(failure.__rich__())

            yield from trajectory_summary(task.formalizations)
            # yield from formalizations_view(task.formalizations)

            if task.hitl_qas:
                yield Static(
                    Text.assemble(
                        ('Human-in-the-Loop Interactions: ', 'bold'),
                        (str(len(task.hitl_qas)), 'bright_cyan'),
                    )
                )

            if task.metadata:
                yield Static(Text('Metadata: ', 'bold'))
                yield Pretty(task.metadata)

        # Create panel
        VerticalGroup.__init__(self, *summary_parts())
        self.border_title = 'Formalization Task'


class StepView(VerticalGroup):
    def __init__(self, step, *args, **kwargs):
        VerticalGroup.__init__(self, *args, **kwargs)
        self.step = step

    def compose(self):
        def response(i):
            response_type = i.response_type
            return {
                'Task': lambda: TaskView(i.task),
                'Message': lambda: Pretty(i.message),
            }[response_type]()

        with TabbedContent(initial='response'):
            with TabPane('Command', id='command'):
                yield command_widget(self.step.command.root)
            with TabPane('Response', id='response'):
                yield response(self.step)



#
#   Imandra Inc.
#
#   codelogician/commands/cmd_doc.py
#

from pathlib import Path
from typing import Annotated, Final

import typer
from rich import print as printr

from codelogician.doc.utils.api_ref import _load_iar

# =====
# # app
# =====

APP_HELP: Final = """\
A CLI tool for working with ImandraX documentation, api reference, examples and other data.

- Guides - articles on writing IML and using ImandraX
- Applications - tutorials on formalizing various domains (e.g. web development) with ImandraX and reasoning about resulting models
- Examples - a set of worked IML examples
- Ref - ImandraX reference (built-in modules and APIs)
- Errors - a collection of
- Approximations - a collection of function approximations that may be useful in your coding

Tips:
- Start with `codelogician doc overview` to see available topics
- `codelogician doc prompt` will generate a prompt you could `feed` into your agent/LLM that has a summary of

- Refer to doc when working with .iml files
    - See available modules: `codelogician doc ref summary`
    - Look up specific modules: `codelogician doc ref list -m <module>`

    - To view
- To dump docs to markdown files: `codelogician doc dump PATH_TO_STORE_DOCS`
- To search the documents: `codelogician doc search TOPIC`
"""

app = typer.Typer(help=APP_HELP, add_completion=False, rich_markup_mode='rich')


# =====
# ### doc overview
# =====


def load_doc():
    from pathlib import Path

    from codelogician.doc.utils.docs import DocManager

    doc_location = Path(__file__).parent / '../doc/data'
    return DocManager(str(doc_location.resolve()))


@app.command(name='contents', help='List the contents of the available documentation')
def contents():
    """
    Show the table of contents
    """

    doc = load_doc()
    doc.print_contents()


@app.command(
    name='prompt', help='Generate a prompt to teach LLMs/Agents how to use ImandraX'
)
def run_prompt_cmd():
    """
    Show IML overview documentation.
    """

    from codelogician.doc.utils.prompts import (
        iml_caveats,
        iml_intro,
        lang_agnostic_meta_eg_overview,
        lang_agnostic_meta_egs_str,
    )

    iml_overview = (
        iml_intro
        + iml_caveats
        + lang_agnostic_meta_eg_overview
        + lang_agnostic_meta_egs_str
    )
    typer.echo(iml_overview)


@app.command(help='Dump IML overview and API reference to markdown files')
def dump(
    dir_path: Annotated[
        Path, typer.Argument(help='Directory path to write markdown files')
    ],
):
    from codelogician.doc.utils.api_ref import (
        dump_guide_docs,
        dump_iml_overview_and_api_refs,
    )

    if not dir_path.exists():
        dir_path.mkdir(exist_ok=False, parents=True)

    dump_iml_overview_and_api_refs(dir_path)
    dump_guide_docs(dir_path)
    typer.secho('Successfully ', fg='green')


@app.command(name='search', help='Search the entire directory')
def search(
    query: Annotated[str, typer.Argument(help='Search query')],
    topic: Annotated[
        str,
        typer.Argument(
            help='Search a specific topic. Must be one of: guide, api, examples, errors, applications'
        ),
    ] = 'all',
):
    """
    Search the documentation with an optional topic parameter.
    """

    if topic.lower() not in [
        'all',
        'guide',
        'api',
        'examples',
        'errors',
        'applications',
    ]:
        typer.secho(
            'Optional parameter `topic` has invalid value. Expected one of: `guide`, `api`, `examples`, `errors`, or `applications`'
        )
        typer.Exit(1)

    doc = load_doc()

    results = doc.find_content(query)
    for r in results:
        printr(r)


@app.command(name='view', help='View a specific chapter or topic in documentation')
def view(
    desc: Annotated[
        str,
        typer.Argument(
            help='View a specific topic. Will look for the closest match (within reason)'
        ),
    ],
):
    doc = load_doc()

    doc_file = doc.find_topic(desc)

    if doc_file is None:
        printr('Found no topic matching the title')
    else:
        printr(doc_file)


# =====
# ### doc api-reference
# =====


api_ref_app = typer.Typer(help='IML Reference (built-in functions, types and modules)')
app.add_typer(api_ref_app, name='ref')

# @api_ref_app.callback(invoke_without_command=True)
# def default(ctx: typer.Context):
#  typer.echo(ctx.get_help())


@api_ref_app.command(
    name='summary', help='Provide an overview of the available modules'
)
def api_ref_summary():
    """
    Show summary of IML Reference (unique modules).
    """

    iar = _load_iar()
    modules = sorted({entry.module for entry in iar})

    typer.echo(
        f'IML Reference - {len(iar)} total entries across {len(modules)} modules\n'
    )
    typer.echo('Available modules:')
    for module in modules:
        count = sum(1 for entry in iar if entry.module == module)
        if module == '':
            typer.echo(f'  (global): {count} entries')
        else:
            typer.echo(f'  {module}: {count} entries')


@api_ref_app.command(
    name='view', help='List IML API reference entries (optionally filtered by module)'
)
def api_ref_view(
    module: Annotated[str | None, typer.Argument(help='Filter by module name')] = None,
):
    """
    List IML API reference entries (optionally filtered by module)."""
    iar = _load_iar()

    if module is not None:
        iar = [entry for entry in iar if entry.module.lower() == module.lower()]
        if not iar:
            typer.secho(
                f'No entries found for module: {module}', fg=typer.colors.RED, err=True
            )
            raise typer.Exit(1)

    typer.echo(f'Showing {len(iar)} entries\n')

    for entry in iar:
        typer.echo(f'Module: {entry.module or "(global)"}')
        typer.echo(f'Name: {entry.name}')
        typer.echo(f'Signature: {entry.signature}')
        if entry.doc:
            typer.echo(f'Doc: {entry.doc}')
        if entry.pattern:
            typer.echo(f'Pattern: {entry.pattern}')
        typer.echo()



#
#   Imandra Inc.
#
#   utils.py
#

from typing import cast

from .proto_models import Error, ErrorMessage, EvalRes


def _error_msg_to_llm_context(
    error_msg: ErrorMessage,
    max_backtrace_len: int = 0,
) -> str:
    locs = error_msg.locs if error_msg.locs else []
    loc_strs: list[str] = [
        (f'({loc.start.line}, {loc.start.col}) - ({loc.stop.line}, {loc.stop.col})')
        for loc in locs
        if (loc.start is not None and loc.stop is not None)
    ]
    loc_str = '; '.join(loc_strs) if loc_strs else ''

    res = f'{error_msg.msg}'
    if loc_str:
        res += f'\nlocs: {loc_str}'
    if error_msg.backtrace and max_backtrace_len > 0:
        res += f'\nbacktrace: {error_msg.backtrace[:max_backtrace_len]}'
    return res


def error_to_llm_context(error: Error, max_stack_depth: int = 3) -> str:
    err_kind = error.kind
    top_msg: str | None = None
    if error.msg is not None:
        top_msg = _error_msg_to_llm_context(error.msg)
    stack_strs = (
        [_error_msg_to_llm_context(msg) for msg in error.stack[:max_stack_depth]]
        if error.stack
        else []
    )
    stack_str = '\n'.join(stack_strs)

    s = ''
    s += f'{top_msg}' if top_msg else ''
    s += f'\n<kind>{err_kind}</kind>' if err_kind else ''
    s += f'\n<stack>\n{stack_str}\n</stack>' if stack_str else ''

    return s


def eval_res_errors_to_llm_context(
    eval_res: EvalRes, max_errors: int = 3
) -> str | None:
    po_errors = eval_res.po_errors
    if not eval_res.errors and not po_errors:
        return None

    errs: list[str] = [error_to_llm_context(err) for err in eval_res.errors]
    po_errs = [error_to_llm_context(err) for err in po_errors]

    res = ''
    for i, err in enumerate(errs[:max_errors], 1):
        res += f'<error_{i}>\n{err}\n</error_{i}>\n\n'
    if po_errors:
        res += 'Proof obligation errors (including termination proving errors):\n\n'
    for i, err in enumerate(po_errs[:max_errors], 1):
        res += f'<po_error_{i}>\n{err}\n</po_error_{i}>\n\n'
    return res


def eval_res_to_llm_context(eval_res: EvalRes) -> str:
    if not eval_res.has_errors:
        s = 'Success!'
        if eval_res.eval_results:
            s += '\n'
        for i, eval_result in enumerate(eval_res.eval_results, 1):
            s += f'\nEval result #{i}:\n'
            s += f'- success: {eval_result.success}\n'
            s += f'- value as ocaml: {eval_result.value_as_ocaml}\n'
            if eval_result.errors:
                s += 'errors:\n'
                for err in eval_result.errors:
                    s += f'- {error_to_llm_context(err)}\n'
        return s

    else:
        s = ''
        s += 'Evaluation errors:\n\n'
        s += cast(str, eval_res_errors_to_llm_context(eval_res))
        return s



#
#   Imandra Inc.
#
#   codelogician/commands/server/cmd_strategy.py
#

from __future__ import annotations

import json
import urllib.parse
from typing import Any

import httpx
import typer

app = typer.Typer(name='strategy', help='CLI for CodeLogician strategy endpoints.')


# -------------------------
# Helpers
# -------------------------


def _print_json(data: Any) -> None:
    typer.echo(json.dumps(data, indent=2, ensure_ascii=False))


def _handle_http_error(resp: httpx.Response) -> None:
    try:
        payload = resp.json()
    except Exception:
        payload = None

    if isinstance(payload, dict) and 'detail' in payload:
        raise typer.BadParameter(f'HTTP {resp.status_code}: {payload["detail"]}')
    raise typer.BadParameter(f'HTTP {resp.status_code}: {resp.text}')


def _get(
    base_url: str,
    path: str,
    *,
    params: dict[str, Any] | None = None,
    timeout: float,
) -> Any:
    url = base_url.rstrip('/') + path
    with httpx.Client(timeout=timeout) as client:
        resp = client.get(url, params=params)
    if resp.status_code >= 400:
        _handle_http_error(resp)
    return resp.json()


def _post(
    base_url: str,
    path: str,
    *,
    params: dict[str, Any] | None = None,
    json_body: Any | None = None,
    timeout: float,
) -> Any:
    url = base_url.rstrip('/') + path
    with httpx.Client(timeout=timeout) as client:
        resp = client.post(url, params=params, json=json_body)
    if resp.status_code >= 400:
        _handle_http_error(resp)

    ctype = resp.headers.get('content-type', '')
    if 'application/json' in ctype:
        return resp.json()
    return resp.text


def _patch(
    base_url: str,
    path: str,
    *,
    json_body: Any | None = None,
    timeout: float,
) -> Any:
    url = base_url.rstrip('/') + path
    with httpx.Client(timeout=timeout) as client:
        resp = client.patch(url, json=json_body)
    if resp.status_code >= 400:
        _handle_http_error(resp)

    ctype = resp.headers.get('content-type', '')
    if 'application/json' in ctype:
        return resp.json()
    return resp.text


def _cfg(ctx: typer.Context) -> dict[str, Any]:
    root = ctx.find_root()
    assert isinstance(root.obj, dict)
    return root.obj


# -------------------------
# Strategy selection / lifecycle
# -------------------------


@app.command('setcws')
def set_cws(
    ctx: typer.Context,
    strat_id: str = typer.Argument(
        ..., help='Strategy ID to set as current working strategy.'
    ),
) -> None:
    """POST /strategy/setcws/{strat_id}"""
    cfg = _cfg(ctx)
    sid = urllib.parse.quote(strat_id, safe='')
    out = _post(cfg['base_url'], f'/strategy/setcws/{sid}', timeout=cfg['timeout'])
    typer.echo(out if isinstance(out, str) else json.dumps(out))


@app.command('cws')
def get_cws(ctx: typer.Context) -> None:
    """GET /strategy/cws"""
    cfg = _cfg(ctx)
    data = _get(cfg['base_url'], '/strategy/cws', timeout=cfg['timeout'])
    if cfg['json_out']:
        _print_json(data)
    else:
        typer.echo('' if data is None else str(data))


@app.command('list')
def list_strategies(ctx: typer.Context) -> None:
    """GET /strategy/list"""
    cfg = _cfg(ctx)
    data = _get(cfg['base_url'], '/strategy/list', timeout=cfg['timeout'])
    if cfg['json_out']:
        _print_json(data)
        return

    # Expect list[{'strat_id':..., 'path':..., 'type':...}]
    if isinstance(data, list):
        for item in data:
            if isinstance(item, dict):
                typer.echo(
                    f'{item.get("strat_id")}  {item.get("type")}  {item.get("path")}'
                )
            else:
                typer.echo(str(item))
    else:
        typer.echo(str(data))


@app.command('states')
def all_states(ctx: typer.Context) -> None:
    """GET /strategy/states"""
    cfg = _cfg(ctx)
    data = _get(cfg['base_url'], '/strategy/states', timeout=cfg['timeout'])
    _print_json(data) if cfg['json_out'] else typer.echo(str(data))


@app.command('create')
def create_strategy(
    ctx: typer.Context,
    strat_type: str = typer.Option(..., '--type', help='Strategy type (e.g., PyIML).'),
    strat_path: str = typer.Option(
        ..., '--path', help='Directory path for this strategy.'
    ),
) -> None:
    """
    POST /strategy/create?strat_type=...&strat_path=...

    Note: endpoint signature is `create_strategy(strat_type: str, strat_path: str)`
    (no Body annotation), so FastAPI typically treats these as query parameters.
    """
    cfg = _cfg(ctx)
    out = _post(
        cfg['base_url'],
        '/strategy/create',
        params={'strat_type': strat_type, 'strat_path': strat_path},
        timeout=cfg['timeout'],
    )
    _print_json(out) if cfg['json_out'] else typer.echo(str(out))


@app.command('delete')
def delete_strategy(
    ctx: typer.Context,
    strat_id: str = typer.Argument(..., help='Strategy ID to delete.'),
) -> None:
    """POST /strategy/delete/{strat_id}"""
    cfg = _cfg(ctx)
    sid = urllib.parse.quote(strat_id, safe='')
    out = _post(cfg['base_url'], f'/strategy/delete/{sid}', timeout=cfg['timeout'])
    typer.echo(out if isinstance(out, str) else json.dumps(out))


# -------------------------
# Current strategy data
# -------------------------


@app.command('state')
def get_state(ctx: typer.Context) -> None:
    """GET /strategy/state"""
    cfg = _cfg(ctx)
    data = _get(cfg['base_url'], '/strategy/state', timeout=cfg['timeout'])
    _print_json(data) if cfg['json_out'] else typer.echo(str(data))


@app.command('summary')
def get_summary(ctx: typer.Context) -> None:
    """GET /strategy/summary"""
    cfg = _cfg(ctx)
    data = _get(cfg['base_url'], '/strategy/summary', timeout=cfg['timeout'])
    _print_json(data) if cfg['json_out'] else typer.echo(str(data))


@app.command('config')
def get_config(ctx: typer.Context) -> None:
    """GET /strategy/config"""
    cfg = _cfg(ctx)
    data = _get(cfg['base_url'], '/strategy/config', timeout=cfg['timeout'])
    _print_json(data) if cfg['json_out'] else typer.echo(str(data))


@app.command('config-set')
def config_set(
    ctx: typer.Context,
    update_json: str = typer.Option(
        ...,
        '--update-json',
        help='JSON object matching StratConfigUpdate; only provided fields will be set.',
    ),
) -> None:
    """
    PATCH /strategy/config/set

    Body: StratConfigUpdate (partial update model).
    """
    cfg = _cfg(ctx)
    try:
        obj = json.loads(update_json)
    except json.JSONDecodeError as e:
        raise typer.BadParameter(f'--update-json must be valid JSON: {e}') from e

    out = _patch(
        cfg['base_url'],
        '/strategy/config/set',
        json_body=obj,
        timeout=cfg['timeout'],
    )
    typer.echo(out if isinstance(out, str) else json.dumps(out, indent=2))


if __name__ == '__main__':
    app()



from dataclasses import dataclass, field
from datetime import datetime
from typing import List


@dataclass
class User:
    id: int
    name: str
    is_driver: bool = False


@dataclass
class Ride:
    id: int
    driver_id: int
    origin: str
    destination: str
    departure: datetime
    seats: int
    booked: List[int] = field(default_factory=list)  # list of rider_ids

    def seats_available(self) -> int:
        return self.seats - len(self.booked)


@dataclass
class Booking:
    id: int
    ride_id: int
    rider_id: int
    booked_at: datetime
    cancelled: bool = False



#
#   Imandra Inc.
#
#   prompts.py
#

short_iml_101 = """IML (Imandra Modeling Language)
- syntax: a formalized Higher-Order subset of OCaml extended with theorem proving \
tactics and verification annotations
- features: counter-example generation, automated induction, conditional rewriting, \
bounded and unbounded verification, theorem proving, model checking, lemma discovery
- logic: Computational Higher-Order Logic (HOL) based on a pure subset of OCaml. \
Recursion and Induction based on ordinals up to ε_0.
- tags: #reasoner #automated reasoning engine #interactive theorem prover \
#automated theorem prover #functional programming language #formal verification
"""

role_prompt = """<role>
You are a software engineer with expertise in OCaml, functional programming, and \
formal semantics. You are tasked to generate code in Imandra Modeling Language (IML) \
either by translating given source code or by generating code from scratch.
</role>\
"""

iml_intro = """
<instructions>
IML is a "pure" (no side effects) subset of OCaml and will be used as input language \
to an automated reasoning engine. Thus, IML code has mechanized formal semantics and \
represents some kind of formalism, making it different from regular OCaml code where \
some side effects are allowed.

We will introduce IML by emphasizing its difference from OCaml and its unique features
as an input language to an automated reasoning engine.

# Imandra Modeling Language (IML): An Introduction

## 1. Core Differences Between IML and OCaml

### 1.1 Standard Library Differences

OCaml standard library modules are either **unavailable** or have **significantly \
different signatures** in IML:

- **Example**: `List.nth` in OCaml has signature `'a list -> int -> 'a`, while in IML \
it's `int -> 'a list -> 'a option`.
  - Parameter order is reversed (index first, then list)
  - Return type uses `option` since IML is pure and cannot raise exceptions for \
invalid indices

**Available modules** in IML include: `Int`, `LChar`, `LString`, `List`, `Map`, \
`Multiset`, `Option`, `Real`, `Result`, `Set`, and `String` — all with IML-specific \
signatures. We'll go through the signatures of these modules in the next section.

### 1.2 Numerical Representation and Precision

IML uses arbitrary precision arithmetic with different numerical type interpretations:

- Literal `3.14` in IML has type `real` (not `float` as in OCaml)
- For type conversion, use:
  - `Real.of_int : int -> real`
  - `Real.to_int : real -> int`

### 1.3 Error Handling Approach

IML is a pure language with no exceptions:

- The `failwith` function available in OCaml **cannot be used** in IML
- Instead, either:
  - Transform partial functions into total functions
  - Use monadic error handling with `Result` or `Option` modules

### 1.4 Type System Constraints

Unlike OCaml, IML restricts function representation in composite types:

- Functions **cannot** be part of algebraic data types, records, or tuples
- For state transition modeling, define:
  - A dedicated event type (with parameterized constructors if needed)
  - A step function that applies events to states

## 2. IML as an Automated Reasoning Engine Input Language

Here's the improved section on termination proving that includes information about \
the Ordinal module:

### 2.1 Termination Proving

Since IML serves as both a programming language and a logic, function termination \
must be provable:

- For common recursion patterns, termination is proven automatically
- For complex cases, provide explicit termination measures using `[@@measure ...]`

**Ordinals in Termination Proofs:**
Termination measures in IML must return values of type `Ordinal.t`. This type \
represents ordinals up to ε₀ in Cantor Normal Form, providing a well-founded ordering \
essential for proving termination. When defining measures, you can use functions like \
`Ordinal.of_int` to convert integers to ordinals, or construct more complex ordinals \
for nested recursion.

**Example with explicit measure:**
```iml
let left_pad_measure n xs =
  Ordinal.of_int (n - List.length xs)

let rec left_pad c n xs =
  if List.length xs >= n then
    xs
  else
    left_pad c n (c :: xs)
[@@measure left_pad_measure n xs]
```

The measure shows that `n - List.length xs` decreases with each recursive call and \
remains non-negative, proving termination. The `Ordinal.of_int` function converts this \
integer difference to an ordinal value that Imandra can use to establish a \
well-founded ordering.

Note that `[@@measure ...]` can only be used for top-level functions, i.e., functions \
that are not defined within other functions.

As a **last resort** when termination is difficult to prove, use `[@@no_validate]`:

```iml
let rec left_pad c n xs =
  if List.length xs >= n then
    xs
  else
    left_pad c n (c :: xs)
[@@no_validate]
```

**Important limitation**: Using `[@@no_validate]` significantly reduces Imandra's \
reasoning capabilities for this function.

### 2.2 Opaque Functions

Opaque functions allow you to define function signatures without implementing their \
behavior in detail. Think of them as mocking funcitonalities that help we pass \
type checking.

#### 2.2.1 Definition

To create an opaque function, attach the `[@@opaque]` attribute to a function with an \
explicit type signature on the function name:

```iml
let sqrt: real -> real = () [@@opaque]
```

#### 2.2.2 Important Syntax Rules

1. **Type annotation placement**: The type annotation must be placed on the function \
name itself, not on parameters or the function body:
   - ✅ `let sqrt: real -> real = () [@@opaque]`
   - ❌ `let sqrt x: real -> real = () [@@opaque]`

2. **Unit**: The implementation body is a unit value `()` no matter what the type is.

#### 2.2.3 When to Use Opaque Functions

Opaque functions are valuable when:
- Working with functions that aren't yet expressible in IML (e.g., operations \
  involving irrational numbers, bitwise operations, etc.)
- Integrating with external third-party libraries
</instructions>
"""

flatten_iml_api_reference = """
<detailed IML modules>
Now we'll go through the signatures of Prelude (not need to import) and modules.

<standard_library>
Prelude (readily available, no need to prefix with module name)
  val || : bool -> bool -> bool
  val && : bool -> bool -> bool
  val = : 'a -> 'a -> bool
  val <> : 'a -> 'a -> bool
  val not : bool -> bool
  val ==> : bool -> bool -> bool
  val <== : bool -> bool -> bool
  val <==> : bool -> bool -> bool
  val + : int -> int -> int
  val const : 'a -> 'b -> 'a
  val >= : int -> int -> bool
  val mk_nat : int -> int
  val < : int -> int -> bool
  val <= : int -> int -> bool
  val > : int -> int -> bool
  val min : int -> int -> int
  val max : int -> int -> int
  val <. : real -> real -> bool
  val <=. : real -> real -> bool
  val >. : real -> real -> bool
  val >=. : real -> real -> bool
  val min_r : real -> real -> real
  val max_r : real -> real -> real
  val ~- : int -> int
  val abs : int -> int
  val - : int -> int -> int
  val ~+ : int -> int
  val * : int -> int -> int
  val / : int -> int -> int
  val mod : int -> int -> int
  val compare : int -> int -> int  (*  Total order, if x = y then 0 else if x < y \
then -1 else 1  *)

type list =
  | []
  | ( :: ) of 'a * 'a list

type result =
  | Ok of 'a
  | Error of 'b
 (*  Result type, representing either a successul result [Ok x] or an error
     [Error x].  *)
</standard_library>

<module_Set>
Module: `Set`
```iml
module type Set = sig
  type Set.t = ('a, bool) Map.t
  val empty : ('a, bool) Map.t
  val full : ('a, bool) Map.t
  val is_empty : 'a Set.t -> bool
  val is_valid : 'a Set.t -> bool
  val mem : 'a -> 'a Set.t -> bool
  val subset : 'a Set.t -> 'a Set.t -> bool
  val add : 'a -> 'a Set.t -> 'a Set.t
  val remove : 'a -> 'a Set.t -> 'a Set.t
  val inter : 'a Set.t -> 'a Set.t -> 'a Set.t
  val union : 'a Set.t -> 'a Set.t -> 'a Set.t
  val complement : 'a Set.t -> 'a Set.t
  val diff : 'a Set.t -> 'a Set.t -> 'a Set.t
  val of_list : 'a list -> ('a, bool) Map.t
  val (++) : 'a Set.t -> 'a Set.t -> 'a Set.t
  val (--) : 'a Set.t -> 'a Set.t -> 'a Set.t
end
```
</module_Set>

<module_Result>
Module: `Result`
```iml
module type Result = sig
  type ('a, 'b) t = ('a, 'b) result
  val return : 'a -> ('a, 'b) result
  val fail : 'a -> ('b, 'a) result
  val map : ('b -> 'c) -> ('b, 'a) result -> ('c, 'a) result
  val map_err : ('a -> 'c) -> ('b, 'a) result -> ('b, 'c) result
  val get_or : default:'a -> ('a, 'b) result -> 'a
  val map_or : default:'a -> ('c -> 'a) -> ('c, 'b) result -> 'a
  val (>|=) : ('b, 'a) result -> ('b -> 'c) -> ('c, 'a) result
  val flat_map : ('b -> ('c, 'a) result) -> ('b, 'a) result -> ('c, 'a) result
  val (>>=) : ('b, 'a) result -> ('b -> ('c, 'a) result) -> ('c, 'a) result
  val fold : ('b -> 'c) -> ('a -> 'c) -> ('b, 'a) result -> 'c
  val is_ok : ('a, 'b) result -> bool
  val is_error : ('a, 'b) result -> bool
  val monoid_product : ('a, 'b) result -> ('c, 'b) result -> ('a * 'c, 'b) result
  val let+ : ('c, 'a) result
  val and+ : (('a * 'c), 'b) result
  val let* : ('c, 'a) result
  val and* : (('a * 'c), 'b) result
end
```
</module_Result>

<module_Real>
Module: `Real`
```iml
module type Real = sig
  type t = real
  val of_int : int -> real
  val _to_int_round_down : real -> int
  val to_int : real -> int
  val (+) : real -> real -> real
  val (-) : real -> real -> real
  val (~-) : real -> real
  val (*) : real -> real -> real
  val (/) : real -> real -> real
  val (<) : real -> real -> bool
  val (<=) : real -> real -> bool
  val (>) : real -> real -> bool
  val (>=) : real -> real -> bool
  val abs : real -> real
  val min : real -> real -> real
  val max : real -> real -> real
  val of_float : float -> real
  val pow : real -> int -> real
end
```
</module_Real>

<module_Option>
Module: `Option`
```iml
module type Option = sig
  type 'a t = 'a option
  val map : ('a -> 'b) -> 'a option -> 'b option
  val map_or : default:'a -> ('b -> 'a) -> 'b option -> 'a
  val is_some : 'a option -> bool
  val is_none : 'a option -> bool
  val return : 'a -> 'a option
  val (>|=) : 'a option -> ('a -> 'b) -> 'b option
  val flat_map : ('a -> 'b option) -> 'a option -> 'b option
  val (>>=) : 'a option -> ('a -> 'b option) -> 'b option
  val or_ : 'a option -> 'a option -> 'a option
  val (<+>) : 'a option -> 'a option -> 'a option
  val exists : ('a -> bool) -> 'a option -> bool
  val for_all : ('a -> bool) -> 'a option -> bool
  val get_or : default:'a -> 'a option -> 'a
  val fold : ('a -> 'b -> 'a) -> 'a -> 'b option -> 'a
  val (<$>) : ('a -> 'b) -> 'a option -> 'b option
  val monoid_product : 'a option -> 'b option -> ('a * 'b) option
  val let+ : 'a option -> ('a -> 'b) -> 'b option
  val and+ : 'a option -> 'b option -> ('a * 'b) option
  val let* : 'a option -> ('a -> 'b option) -> 'b option
  val and* : 'a option -> 'b option -> ('a * 'b) option
end
```
</module_Option>

<module_Int>
Module: `Int`
```iml
module type Int = sig
  type t = int
  val (+) : int -> int -> int
  val (-) : int -> int -> int
  val (~-) : int -> int
  val (*) : int -> int -> int
  val (/) : int -> int -> int
  val (mod) : int -> int -> int
  val (<) : int -> int -> bool
  val (<=) : int -> int -> bool
  val (>) : int -> int -> bool
  val (>=) : int -> int -> bool
  val min : int -> int -> int
  val max : int -> int -> int
  val abs : int -> int
  val to_string : int -> string
  val compare : int -> int -> int
  val equal : 'a -> 'a -> bool
  val pow : int -> int -> int
  val mod_zero_prod : int -> int -> int -> bool
  val mod_sub_id : int -> int -> bool
end
```
</module_Int>

<module_List>
Module: `List`
```iml
module type List = sig
  type 'a t = 'a list
  val empty : 'a list
  val is_empty : 'a list -> bool
  val cons : 'a -> 'a list -> 'a list
  val return : 'a -> 'a list
  val hd : 'a list -> 'a
  val tl : 'a list -> 'a list
  val head_opt : 'a list -> 'a option
  val append : 'a list -> 'a list -> 'a list
  val append_to_nil : 'a list -> bool
  val append_single : 'a -> 'a list -> 'a list -> bool
  val rev : 'a list -> 'a list
  val length : 'a list -> int
  val len_nonnegative : 'a list -> bool
  val len_zero_is_empty : 'a list -> bool
  val len_append : 'a list -> 'a list -> bool
  val split : ('a * 'b) list -> 'a list * 'b list
  val map : ('a -> 'b) -> 'a list -> 'b list
  val map2 : ('c -> 'a -> 'b) -> 'c list -> 'a list -> ('b list, string) result
  val for_all : ('a -> bool) -> 'a list -> bool
  val exists : ('a -> bool) -> 'a list -> bool
  val fold_left : ('b -> 'a -> 'b) -> 'b -> 'a list -> 'b
  val fold_right : ('b -> 'a -> 'a) -> 'b list -> 'a -> 'a
  val mapi : (int -> 'b -> 'a) -> 'b list -> 'a list
  val filter : ('a -> bool) -> 'a list -> 'a list
  val filter_map : ('a -> 'b option) -> 'a list -> 'b list
  val flat_map : ('b -> 'a list) -> 'b list -> 'a list
  val find : ('a -> bool) -> 'a list -> 'a option
  val mem : 'a -> 'a list -> bool
  val mem_assoc : 'a -> ('a * 'b) list -> bool
  val nth : int -> 'a list -> 'a option
  val assoc : 'a -> ('a * 'b) list -> 'b option
  val bounded_recons : int -> 'a list -> 'a list
  val take : int -> 'a list -> 'a list
  val drop : int -> 'a list -> 'a list
  val range : int -> int -> int list
  val (--) : int -> int -> int list
  val insert_sorted : leq:('a -> 'a -> bool) -> 'a -> 'a list -> 'a list
  val sort : leq:('a -> 'a -> bool) -> 'a list -> 'a list
  val is_sorted : leq:('a -> 'a -> bool) -> 'a list -> bool
  val monoid_product : 'a list -> 'b list -> ('a * 'b) list
  val (>|=) : 'a list -> ('a -> 'b) -> 'b list
  val (>>=) : 'b list -> ('b -> 'a list) -> 'a list
  val let+ : 'b list -> ('b -> 'a) -> 'a list
  val and+ : 'a list -> 'b list -> ('a * 'b) list
  val let* : 'b list -> ('b -> 'a list) -> 'a list
  val and* : 'a list -> 'b list -> ('a * 'b) list
end
```
</module_List>

<module_Map>
Module: `Map`
```iml
module type Map = sig
  type Map.t = {{| l : ('a * 'b) list; | default : 'b}}
  val const : 'b -> ('a, 'b) Map.t
  val add' : ('a, 'b) Map.t -> 'a -> 'b -> ('a, 'b) Map.t
  val add : 'a -> 'b -> ('a, 'b) Map.t -> ('a, 'b) Map.t
  val get_default : ('a, 'b) Map.t -> 'b
  val get' : ('a, 'b) Map.t -> 'a -> 'b
  val get : 'a -> ('a, 'b) Map.t -> 'b
  val of_list : 'b -> ('a * 'b) list -> ('a, 'b) Map.t
end
```
</module_Map>

<module_Multiset>
Module: `Multiset`
```iml
module type Multiset = sig
  type Multiset.t = ('a, int) Map.t
  val empty : ('a, int) Map.t
  val add : 'a -> ('a, int) Map.t -> ('a, int) Map.t
  val find : 'a -> ('a, int) Map.t -> int
  val mem : 'a -> ('a, int) Map.t -> bool
  val remove : 'a -> ('a, int) Map.t -> ('a, int) Map.t
  val of_list : 'a list -> ('a, int) Map.t
end
```
</module_Multiset>

<module_LChar>
In IML, we mainly use `LChar` and `LString`, (logic-mode character and logic-mode\
string). To create a logic-mode string (LString.t), you can use the syntax sugar\
`{{l|hello|l}}`.

```iml repl
> let y = {{l|hello|l}};;
val y: LChar.t list
```
</module_LChar>

<module_LChar>
Module: `LChar`
```iml
module type LChar = sig
  type LChar.t =
 LChar.Char of bool * bool * bool * bool * bool * bool * bool * bool
  val zero : LChar.t
  val is_printable : LChar.t -> bool
end
```
</module_LChar>

<module_LString>
Module: `LString`
```iml
module type LString = sig
  type t = LChar.t list
  val empty : LChar.t list
  val of_list : 'a -> 'a
  val length : LChar.t list -> int
  val len_pos : LChar.t list -> bool
  val len_zero_inversion : LChar.t list -> bool
  val append : LChar.t list -> LChar.t list -> LChar.t list
  val (^^) : LChar.t list -> LChar.t list -> LChar.t list
  val for_all : (LChar.t -> bool) -> LChar.t list -> bool
  val exists : (LChar.t -> bool) -> LChar.t list -> bool
  val concat : LChar.t list -> LChar.t list list -> LChar.t list
  val is_printable : LChar.t list -> bool
  val sub : LChar.t list -> int -> int -> LChar.t list
  val prefix : LChar.t list -> LChar.t list -> bool
  val suffix : LChar.t list -> LChar.t list -> bool
  val contains : LChar.t list -> LChar.t list -> bool
  val take : int -> LString.t -> LString.t
  val drop : int -> LString.t -> LString.t
end
```
</module_LString>

<module_String>
Module: `String`
```iml
module type String = sig
(* Note:
- These strings correspond to OCaml native strings, and do not have a particular \
unicode encoding.
- Rather, they should be seen as sequences of bytes, and it is also this way that \
Imandra considers them.
 *)
  type String.t = string
  val empty : string
  val length : String.t -> int
  val append : String.t -> String.t -> String.t
  val concat : string -> String.t list -> string
  val prefix : String.t -> String.t -> bool
  val suffix : String.t -> String.t -> bool
  val contains : String.t -> String.t -> bool
  val unsafe_sub : String.t -> int -> int -> String.t
  val sub : string -> int -> int -> String.t option
  val of_int : int -> string
  val unsafe_to_nat : String.t -> int
  val to_nat : string -> int option
  val is_nat : string -> bool
  val is_int : string -> bool
  val unsafe_to_int : string -> int
  val to_int : string -> int option
end
```
</module_String>
"""


lang_agnostic_meta_eg_overview = """\
Now, let's see a few educational examples. Each example demonstrates key \
concepts in IML through:

1. Explanatory comments before functions that highlight important concepts
2. REPL evaluation results showing type signatures and values. Comments marked with \
(* val ... *) show the REPL's response, helping you understand type inference and \
evaluation.

These comments are only for educational purposes. They teach you about the thinking \
process of writing IML code but they are not desired to be any generated IML code.\
"""

lang_agnostic_meta_eg_1 = """\
(* Float-like literal is interpreted as type `real`, arbitary precision real number. *)

let pi = 3.14159

(* val pi : real = 314159/100000 *)


(* Arithemetic for real:

- multiplication `*.`

- division `/.`

- etc *)

let circle_area (d : real) : real =

let r = d /. 2.0 in

pi *. r *. r

(* val circle_area : real -> real = <fun> *)\
"""

lang_agnostic_meta_eg_2 = """\
(* Note: `int` is the type for integers. *)
let count_negatives_in_row (row : int list) : int =
    List.fold_left (fun acc num -> if num < 0 then acc + 1 else acc) 0 row
  (* val count_negatives_in_row : int list -> int = <fun> *)

  let x = count_negatives_in_row [-1; -2; -3; 4; 5]
  (* val x : int = 3 *)

  (* Note:
    - `List.nth: int -> 'a list -> 'a option`, NOT returning 'a as in OCaml
    - That's why we need binding operator `let+` and `let*` from `Option` module
   *)
  let count_negative_numbers_in_a_column_wise_row_wise_sorted_matrix
      (m : int list list)
      (n : int)
      (cols : int)
      : int option =
    List.fold_left
      (fun acc i ->
        (* import Option module locally to use `let*`. could also be imported at the\
beginning by `open Option` *)
        let open Option in
        let* acc_val = acc in
        let+ row_count = map count_negatives_in_row (List.nth i m) in
        acc_val + row_count)
      (Some 0)
      (0 -- (n - 1))
  (* val count_negative_numbers_in_a_column_wise_row_wise_sorted_matrix :
    int list list -> int -> int -> int option = <fun> *)
"""

lang_agnostic_meta_eg_3 = """\
(* Note:
- integer division and modulo in IML: `mod: int -> int -> int`. And `/`.
 *)
 let is_divisor (n : int) (divisor : int) : bool =
    n mod divisor = 0
  (* val is_divisor : int -> int -> bool = <fun> *)


  let add_divisors (n : int) (divisor : int) (current_sum : int) : int =
    if divisor = n / divisor then
      current_sum + divisor
    else
      current_sum + divisor + n / divisor
  (* val add_divisors : int -> int -> int -> int = <fun> *)

  let sqrt : real -> real = () [@@opaque]
  (* val sqrt : real -> real = <fun> *)

  (* Note:
  - conversion between `int` and `real`: `Real.of_int` and `Real.to_int`.
   *)
  let sum_of_all_proper_divisors_of_a_natural_number (num : int) : int =
    let calculate_sum_of_divisors (n : int) (limit : int) : int =
      List.fold_left
        (fun acc i ->
          if is_divisor n i then
            acc + add_divisors n i 0
          else
            acc)
        0
        (2 -- limit)
    in
    let limit = Real.to_int (Real.of_int num |> sqrt) in
    calculate_sum_of_divisors num limit + 1
  (* val sum_of_all_proper_divisors_of_a_natural_number : int -> int = <fun> *)
"""

lang_agnostic_meta_eg_4 = """\
(* Another example with conversion between `real` and `int`. *)
let sum_of_first_n_numbers (n: int) : real =
  (Real.of_int n *. (Real.of_int n +. 1.0)) /. 2.0
(* val sum_of_first_n_numbers : int -> real = <fun> *)

let sum_of_squares_of_first_n_numbers (n: int) : real =
  (Real.of_int n *. (Real.of_int n +. 1.0) *. (2.0 *. Real.of_int n +. 1.0)) /. 6.0

(* val sum_of_squares_of_first_n_numbers : int -> real = <fun> *)


let sum_matrix_element_absolute_difference_row_column_numbers_2 (n: int) : int =
  let sum_first_n = sum_of_first_n_numbers n in
  let sum_squares_n = sum_of_squares_of_first_n_numbers n in
  (sum_first_n +. sum_squares_n) |> Real.to_int
(* val sum_matrix_element_absolute_difference_row_column_numbers_2 : int -> int = \
<fun> *)
"""

lang_agnostic_meta_eg_5 = """\
(* Note:
- for opeartions that is unavailable in IML, `[@@opaque]` can be attached
  to a dummy implementation with correct type signature. Such implementation
  can be used to bypass unavailable functions during compilation.
- Example of unavailable operations: bitwise operations (e.g., `land`, `lxor`),
  irrational arithmetic (e.g., square root, exponentiation), random.
 *)
let xor : int -> int -> int = () [@@opaque]
(* val xor : int -> int -> int = <fun> *)

let rec helper state pairs =
  match pairs with
  | [] -> state
  | (a, b) :: rest ->
    let new_state = min state (xor a b) in
    helper new_state rest
(* val helper : int -> (int * int) list -> int = <fun> *)

let find_minimum_xor (arr : int list) : int =
  let initial_state = 999999 in
  let pairs = List.monoid_product arr arr |> List.filter (fun (a, b) -> a < b) in
  helper initial_state pairs
(* val find_minimum_xor : int list -> int = <fun> *)

let minimum_xor_value_pair (arr : int list) : int =
  let sorted_arr = List.sort ~leq:(fun a b -> a <= b) arr in
  find_minimum_xor sorted_arr
(* val minimum_xor_value_pair : int list -> int = <fun> *)
"""

lang_agnostic_meta_eg_6 = """\
(* Another example using `[@@opaque]` *)
open Option

let random_int : int -> int = () [@@opaque]
(* val random_int : int -> int = <fun> *)

let swap_elements lst i j =
  let get_i = List.nth i lst in
  let get_j = List.nth j lst in
  List.mapi (fun idx x ->
    if idx = i then
      get_j |> get_or ~default:x
    else if idx = j then
      get_i |> get_or ~default:x
    else
      x
  ) lst
(* val swap_elements : 'a list -> int -> int -> 'a list = <fun> *)

let shuffle_a_given_array (arr : int list) : int list =
  let rec helper state i =
    if i <= 0 then
      state
    else
      let j = random_int (i + 1) in
      let state = swap_elements state i j in
      helper state (i - 1)
  in
  helper arr (List.length arr - 1)
(* val shuffle_a_given_array : int list -> int list = <fun> *)
"""

lang_agnostic_meta_eg_7 = """\
(* Character and String:
- IML supports logic-mode character `LChar.t` and logic-mode string `LString.t`.
- `{l|...|l}` is used to create a logic-mode string, `LString.t`, aka, `LChar.t list`.
- To create a logic-mode character, `LChar.t`, you can use list operations on \
`LString.t`, eg `List.hd {l|...|l}`.
 *)
let char_0: LChar.t = List.hd {l|0|l}
let char_1: LChar.t = List.hd {l|1|l}
(* val char_0 : LChar.t = '0'
val char_1 : LChar.t = '1' *)

let count_bits (s : LString.t) : int * int =
  let zeros = List.fold_left (fun acc ch -> if ch = char_0 then acc + 1 else acc) 0 s in
  let ones = List.length s - zeros in
  (zeros, ones)
(* val count_bits : LString.t -> int * int = <fun> *)

let change_bits_can_made_one_flip (s : LString.t) : bool =
  let zeros, ones = count_bits s in
  zeros = 1 || ones = 1
(* val change_bits_can_made_one_flip : LString.t -> bool = <fun> *)
"""


lang_agnostic_meta_egs: list[str] = [
    lang_agnostic_meta_eg_1,
    lang_agnostic_meta_eg_2,
    lang_agnostic_meta_eg_3,
    lang_agnostic_meta_eg_4,
    lang_agnostic_meta_eg_5,
    lang_agnostic_meta_eg_6,
    lang_agnostic_meta_eg_7,
]

lang_agnostic_meta_egs_str = ''
for i, eg in enumerate(lang_agnostic_meta_egs, 1):
    lang_agnostic_meta_egs_str += f'Example {i}:'
    lang_agnostic_meta_egs_str += '\n\n'
    lang_agnostic_meta_egs_str += '```iml\n' + eg + '\n```\n\n'

iml_101: str = (
    iml_intro
    + '\n\n'
    + flatten_iml_api_reference
    + '\n\n'
    + lang_agnostic_meta_eg_overview
    + '\n\n'
    + lang_agnostic_meta_egs_str
    + '\n\n'
)

iml_caveats = """\
- Prefer pattern matching over using auxiliary functions if possible because it \
makes recursive steps explicit and visible, and directly reflects data's inductive \
structure. In comparison, auxiliary functions like `List.nth`, `List.length` \
add a layer of abstraction that obscures the underlying recursive structure, \
making formal verification more complex.
- Nested recursive functions complicate the proof process. Try to extract the \
nested function into a top-level function if possible.
"""



from __future__ import annotations

from enum import Enum
from typing import Literal, Self

from devtools import pformat
from pydantic import BaseModel, Field, TypeAdapter, field_validator, model_validator

from codelogician.util import raise_

from ..decode_artifact import RegionStr, decode_artifact
from .artmsg import Art
from .error import Error
from .session import Session
from .task import Origin, Task
from .utils import Empty, StringMsg


class SessionCreateReq(BaseModel):
    api_version: str = Field(description='the API types version (mandatory)')


class LiftBool(Enum):
    Default = 0
    NestedEqualities = 1
    Equalities = 2
    All = 3


class DecomposeReq(BaseModel):
    session: Session | None = Field(default=None)
    name: str = Field(description='name of function to decompose')
    assuming: str | None = Field(
        default=None, description='name of side condition function'
    )
    basis: list[str]
    rule_specs: list[str]
    prune: bool
    ctx_simp: bool | None = Field(default=None)
    lift_bool: LiftBool | None = Field(default=None)
    str_: bool = Field(default=True, alias='str')
    timeout: int | None = Field(default=None)


class DecomposeResProto(BaseModel):
    artifact: Art | None = Field(default=None)
    err: Empty | None = Field(default=None)
    errors: list[Error] = Field(default_factory=list)
    task: Task | None = Field(default=None)

    @model_validator(mode='after')
    def one_of_res(self) -> Self:
        if (self.artifact is None) == (self.err is None):
            raise ValueError('One of artifact or err must be set')
        return self


class string_kv(BaseModel):
    k: str
    v: str


class DecomposeRes(DecomposeResProto):
    """Result of a decomposition"""

    regions_str: list[RegionStr] | None = Field(
        default=None, description="None if there's decomposition error"
    )

    @model_validator(mode='after')
    def unwrap_region_str(self) -> Self:
        # Unwrap the regionsStr from the artifact to the old format
        artifact = self.artifact
        # fmt: off
        return (self if self.regions_str is not None or self.errors else
                raise_(ValueError('DecomposeRes requires artifact')) if artifact is None else
                self.model_copy(update={'regions_str': decode_artifact(artifact.data, artifact.kind)}))
        # fmt: on

    def __repr__(self) -> str:
        return pformat(self)

    @property
    def iml_test_cases(self) -> list[dict[str, str]]:
        """
        [
            {"args": {"x": "1", "y": "2"}, "expected_output": "3"},
            {"args": {"x": "3", "y": "4"}, "expected_output": "7"},
        ]
        """
        if not self.regions_str:
            return []
        test_cases = []
        for region_str in self.regions_str:
            if (model_eval_str := region_str.model_eval_str) is None:
                continue
            test_cases.append(
                {'args': region_str.model_str, 'expected_output': model_eval_str}
            )
        return test_cases

    @property
    def test_docstrs(self) -> list[str]:
        docstrs = []
        if not self.regions_str:
            return docstrs
        for region_str in self.regions_str:
            s = ''
            if region_str.constraints_str:
                s += 'Constraints:\n'
                for c in region_str.constraints_str:
                    s += f'    - `{c}`\n'
            if region_str.invariant_str:
                s += 'Invariant:\n'
                s += f'    - `{region_str.invariant_str}`\n'
            docstrs.append(s)
        return docstrs


class EvalSrcReq(BaseModel):
    session: Session | None = Field(default=None)
    src: str = Field(description='source code to evaluate')
    async_only: bool | None = Field(
        default=None,
        description='if true, do not wait for tasks results, only return the task list '
        'and not the task results. Use `get_artifact` to get the results.',
    )


class EvalOutput(BaseModel):
    """Output of an `eval` statement"""

    success: bool
    value_as_ocaml: str | None = Field(
        default=None, description='result as a OCaml value, if any'
    )
    errors: list[Error] = Field(default_factory=list)


class EvalRes(BaseModel):
    success: bool = Field()
    messages: list[str] = Field(default_factory=list, description='"normal" messages')
    errors: list[Error] = Field(default_factory=list, description='akin to stderr')

    # All tasks started during eval
    tasks: list[Task] = Field(
        default_factory=list, description='all tasks started during eval'
    )
    po_results: list[PO_Res] = Field(default_factory=list)
    eval_results: list[EvalOutput] = Field(default_factory=list)
    decomp_results: list[DecomposeRes] = Field(default_factory=list)

    def __repr__(self) -> str:
        return pformat(self, indent=2)

    @property
    def po_errors(self) -> list[Error]:
        return [err for po_res in self.po_results for err in po_res.errors]

    @property
    def has_errors(self) -> bool:
        return (len(self.errors) > 0) or (len(self.po_errors) > 0)


class VerifySrcReq(BaseModel):
    session: Session | None = Field(default=None)
    src: str = Field(description='source code')
    hints: str | None = Field(default=None)


class VerifyNameReq(BaseModel):
    session: Session | None = Field(default=None)
    name: str = Field(description='name of the predicate to verify')
    hints: str | None = Field(default=None)


class InstanceSrcReq(BaseModel):
    session: Session | None = Field(default=None)
    src: str = Field(description='source code')
    hints: str | None = Field(default=None)


class InstanceNameReq(BaseModel):
    session: Session | None = Field(default=None)
    name: str = Field(description='name of the predicate to verify')
    hints: str | None = Field(default=None)


class Proved(BaseModel):
    proof_pp: str | None = Field(default=None)


class Verified_upto(BaseModel):
    msg: str | None = Field(default=None)


class Unsat(BaseModel):
    proof_pp: str | None = Field(default=None)


class ModelType(Enum):
    Counter_example = 'Counter_example'
    Instance = 'Instance'


class Model(BaseModel):
    m_type: ModelType  # NOTE: proto file has this but it's not returned?
    src: str = Field(description='iml source code for the model')
    artifact: Art | None = Field(default=None, description='the model as an artifact')


class Refuted(BaseModel):
    model: Model | None = Field(default=None)


class Sat(BaseModel):
    model: Model | None = Field(default=None)


class CounterSat(BaseModel):
    model: Model | None = Field(default=None)


class PO_Res(BaseModel):
    unknown: StringMsg | None = Field(default=None)
    err: Empty | None = Field(default=None)
    proof: Proved | None = Field(default=None)
    instance: CounterSat | None = Field(default=None)
    verified_upto: Verified_upto | None = Field(default=None)

    errors: list[Error] = Field(default_factory=list)
    task: Task | None = Field(default=None, description='the ID of the task')
    origin: Origin | None = Field(
        default=None, description='where did the task originate?'
    )

    @model_validator(mode='after')
    def one_of_res(self) -> Self:
        sum_of_res = sum(
            1
            for r in [
                self.unknown,
                self.err,
                self.proof,
                self.instance,
                self.verified_upto,
            ]
            if r is not None
        )
        if sum_of_res != 1:
            raise ValueError(
                'Exactly one of unknown, err, proof, instance, verified_upto must be '
                'set'
            )
        return self


class VerifyRes(BaseModel):
    unknown: StringMsg | None = Field(default=None)
    err: Empty | None = Field(default=None)
    proved: Proved | None = Field(default=None)
    refuted: Refuted | None = Field(default=None)
    verified_upto: Verified_upto | None = Field(default=None)

    errors: list[Error] = Field(default_factory=list)
    task: Task | None = Field(default=None, description='the ID of the task')

    @property
    def res_type(
        self,
    ) -> Literal['unknown', 'err', 'proved', 'refuted', 'verified_upto']:
        if self.unknown is not None:
            return 'unknown'
        elif self.err is not None:
            return 'err'
        elif self.proved is not None:
            return 'proved'
        elif self.refuted is not None:
            return 'refuted'
        elif self.verified_upto is not None:
            return 'verified_upto'
        else:
            raise AssertionError('Never')

    @property
    def res(self) -> StringMsg | Empty | Proved | Refuted | Verified_upto:
        if self.unknown is not None:
            return self.unknown
        elif self.err is not None:
            return self.err
        elif self.proved is not None:
            return self.proved
        elif self.refuted is not None:
            return self.refuted
        elif self.verified_upto is not None:
            return self.verified_upto
        else:
            raise AssertionError('Never')

    @model_validator(mode='after')
    def one_of_res(self) -> Self:
        sum_of_res = sum(
            1
            for r in [
                self.unknown,
                self.err,
                self.proved,
                self.refuted,
                self.verified_upto,
            ]
            if r is not None
        )
        if sum_of_res != 1:
            raise ValueError(
                'Exactly one of unknown, err, proved, refuted, verified_upto must be '
                'set'
            )
        return self


class InstanceRes(BaseModel):
    unknown: StringMsg | None = Field(default=None)
    err: Empty | None = Field(default=None)
    unsat: Unsat | None = Field(default=None)
    sat: Sat | None = Field(default=None)

    errors: list[Error] = Field(default_factory=list)
    task: Task | None = Field(default=None, description='the ID of the task')

    @model_validator(mode='after')
    def one_of_res(self) -> Self:
        sum_of_res = sum(
            1 for r in [self.unknown, self.err, self.unsat, self.sat] if r is not None
        )
        if sum_of_res != 1:
            raise ValueError('Exactly one of unknown, err, unsat, sat must be set')
        return self


class TypecheckReq(BaseModel):
    session: Session | None = Field(default=None)
    src: str = Field(description='source code to evaluate')


class TypecheckResProto(BaseModel):
    success: bool
    types: str = Field(description='JSON string of inferred types')
    errors: list[Error] = Field(default_factory=list)


class InferredType(BaseModel):
    name: str
    ty: str = Field(description='inferred type')
    line: int = Field(description='line number')
    column: int = Field(description='column number')


InferredTypes = TypeAdapter(list[InferredType])


class TypecheckRes(TypecheckResProto):
    types: list[InferredType] = Field(description='Parsed inferred types')

    @field_validator('types', mode='before')
    @classmethod
    def validate_types(cls, v: str | list[InferredType]) -> list[InferredType]:
        if isinstance(v, str):
            # Parse JSON string into list of dicts and validate with pydantic
            types_list = InferredTypes.validate_json(v)
            return types_list
        return v


class OneshotReq(BaseModel):
    input: str = Field(description='some iml code')
    timeout: float | None = Field(default=None)


class OneshotRes(BaseModel):
    class Stats(BaseModel):
        time: float

    results: list[str] = Field(default_factory=list)
    errors: list[str] = Field(default_factory=list)
    stats: Stats | None = Field(default=None)
    detailed_results: list[str] = Field(default_factory=list)



from basic import add, absolute, multiply, subtract, is_positive


def square(x: float) -> float:
    """Calculate square of x."""
    return multiply(x, x)


def cube(x: float) -> float:
    """Calculate cube of x."""
    return multiply(multiply(x, x), x)



def distance_1d(x1: float, x2: float) -> float:
    """Calculate distance between two points on a line."""
    return absolute(subtract(x2, x1))


def average(a: float, b: float) -> float:
    """Calculate average of two numbers."""
    return multiply(add(a, b), 0.5)


def sum_of_squares(a: float, b: float) -> float:
    """Calculate sum of squares of two numbers."""
    return add(square(a), square(b))


def is_negative(x: float) -> bool:
    """Check if number is negative."""
    return not is_positive(x) and x != 0


def sign(x: float) -> float:
    """Return sign of number: 1.0 for positive, -1.0 for negative, 0.0 for zero."""
    if is_positive(x):
        return 1.0
    elif x == 0:
        return 0.0
    else:
        return -1.0


import re
from enum import Enum
from typing import Self

from pydantic import BaseModel, Field

from .locs import Location


# from https://github.com/imandra-ai/imandrax-api/blob/1a4abb4eb59d9545c6f0af9698956437b5b5dcb8/src/internal/error_data/imandrax_errors.ml
class ErrorKind(str, Enum):
    APPLIED_SYMBOL_TYPE_ERR = 'AppliedSymbolTypeErr'
    CONFIG_ERROR = 'ConfigError'
    CPTR_GET_OTHER_ERR = 'CptrGetOtherErr'
    CPTR_NOT_FOUND_IN_STORAGE = 'CptrNotFoundInStorage'
    GENERIC_USER_ERROR = 'GenericUserError'
    INVALID_FUN_DEFINITION = 'InvalidFunDefinition'
    INVALID_QIDENT = 'InvalidQident'
    INVALID_TYPE_ALIAS = 'InvalidTypeAlias'
    INVALID_TYPE_DEFINITION = 'InvalidTypeDefinition'
    INVALID_CNAME = 'InvalidCname'
    LEVAL_ERROR = 'LevalError'
    LOWER_CIR_ERROR = 'LowerCirError'
    LOWER_MIR_ERROR = 'LowerMirError'
    LOWER_RIR_ERROR = 'LowerRirError'
    PARSER_NOT_REGISTERED = 'ParserNotRegistered'
    PATMATCH_ERROR = 'PatmatchError'
    SYN_TERM_CREATE_ERROR = 'SynTermCreateError'
    SYNTAX_ATTRIBUTE_ERR = 'SyntaxAttributeErr'
    SYNTAX_ERR = 'SyntaxErr'
    SYNTAX_UNSUGAR_ERR = 'SyntaxUnsugarErr'
    TACTIC_EVAL_ERR = 'TacticEvalErr'
    TERM_INVALID_SHAPE = 'TermInvalidShape'
    TERM_MODEL_FIERR = 'TermModelFIErr'
    THREAD_FUT_FAILURE = 'ThreadFutFailure'
    THREAD_TIMER_FAILURE = 'ThreadTimerFailure'
    TY_INFER_INTERNAL_ERROR = 'TyInferInternalError'
    TYPE_ARITY_MISMATCH = 'TypeArityMismatch'
    TYPE_CYCLE_DETECTED = 'TypeCycleDetected'
    TYPE_ERR = 'TypeErr'
    TYPE_VAR_ALREADY_BOUND = 'TypeVarAlreadyBound'
    TYPED_SYMBOL_NON_GROUND = 'TypedSymbolNonGround'
    UID_CONTENT_ADDRESSING_NAME_MISMATCH = 'UidContentAddressingNameMismatch'
    UID_CONTENT_ADDRESSING_NOT_TEMPORARY = 'UidContentAddressingNotTemporary'
    UID_NOT_CONTENT_ADDRESSED = 'UidNotContentAddressed'
    UNKNOWN_BUILTIN_SYMBOL_FOR_UID = 'UnknownBuiltinSymbolForUid'
    UNKNOWN_TYPE_DEFINITION = 'UnknownTypeDefinition'
    INTERACTIVE_PROOF_ERR = 'InteractiveProofErr'
    PROOF_DESER_ERROR = 'ProofDeserError'
    PROOF_CHECK_ERROR = 'ProofCheckError'
    INVALID_ANCHOR = 'InvalidAnchor'
    INDUCT_SCHEMA_ERROR = 'InductSchemaError'
    UNSUPPORTED = 'Unsupported'
    LOWER_FOL_ERROR = 'LowerFolError'
    VALIDATION_ERROR = 'ValidationError'
    LSP_ERROR = 'LspError'
    INTERRUPTED = 'Interrupted'
    REDIS_ERROR = 'RedisError'
    CIRDECL_NOT_FOUND_IN_STORAGE = 'CIRDeclNotFoundInStorage'
    SERIALIZATION_ERROR = 'SerializationError'
    DESERIALIZATION_ERROR = 'DeserializationError'
    DEBUG_MODE = 'DebugMode'
    IMPORT_ERROR = 'ImportError'
    GENERIC_IOERROR = 'GenericIOError'
    DUNE_ERROR = 'DuneError'
    RPC_ERROR = 'RpcError'
    RPC_DESER_ERROR = 'RpcDeserError'
    RPC_NETWORK_ERROR = 'RpcNetworkError'
    RPC_TIMEOUT = 'RpcTimeout'
    AUTH_ERROR = 'AuthError'
    FILE_EXISTS = 'FileExists'
    DIRECTORY_CREATION_ERROR = 'DirectoryCreationError'
    DECOMP_ERROR = 'DecompError'
    VERSION_MISMATCH_ERROR = 'VersionMismatchError'
    OH_NO_ERROR = 'OhNoError'
    DEBOUNCED = 'Debounced'

    @classmethod
    def from_proto_kind(cls, proto_kind: str) -> Self:
        kinds = re.findall(r'\{ Kind.name = "(.+)" \}', proto_kind)
        if len(kinds) != 1:
            raise ValueError('Unable to parse kind')
        return cls(kinds[0])


class ErrorMessage(BaseModel):
    """An error message"""

    msg: str
    locs: list[Location] = Field(
        default_factory=list, description='Locations for this message'
    )
    backtrace: str | None = Field(default=None, description='Captured backtrace')


class Error(BaseModel):
    msg: ErrorMessage | None = Field(
        default=None, description='The toplevel error message'
    )
    kind: str = Field(description='A string description of the kind of error')
    stack: list[ErrorMessage] = Field(
        default_factory=list, description='Context for the error'
    )
    process: str | None = Field(default=None)



import datetime
from enum import Enum
from functools import partial
from pathlib import Path

from pydantic import BaseModel
from rich.console import Group
from rich.text import Text
from textual.containers import HorizontalGroup, ScrollableContainer, VerticalGroup
from textual.reactive import reactive
from textual.screen import Screen
from textual.widget import Widget
from textual.widgets import Button, Footer, Static
from textual.widgets._header import HeaderIcon, HeaderTitle

from codelogician.modeling.task_creator import ModelTaskCreator

from ..util import head, maybe, maybe_else

_attribs = {
    'error': ('red', 'Error'),
    'info': ('blue', 'Info'),
    'success': ('green', 'Success'),
}


def bind(screen, *names):
    def update(name, _, new):
        setattr(screen, name, new)

    for name in names:
        screen.watch(screen.app, name, partial(update, name))


class Source(Enum):
    NoSource = 'No_source'
    Server = 'Server'
    Disk = 'Disk'


class Status(BaseModel):
    source: Source

    disk_loaded: bool | None = None
    disk_error: str | None = None
    disk_path: Path | None = None

    server_addr: str | None = 'http://127.0.0.1:8000'
    server_last_update: datetime.datetime | None = None
    server_error: str | None = None

    def error_status(self):
        return (
            ('SERVER ERROR' if self.server_error else '')
            if self.source == Source.Server
            else ('DISK ERROR' if self.disk_error else '')
        )

    # fmt: off
    def __rich__(self):
        def first_line(s): return head(s.splitlines())
        if self.source == Source.Server:
            fields = [('Source: ', 'Server'),
                      ('Server address: ', self.server_addr),
                      ('Server last update: ', self.server_last_update),
                      ('Server error: ', self.server_error)]
        else:
            fields = [('Source: ', 'Disk'),
                      ('Disk path: ', self.disk_path),
                      ('Disk loaded: ', self.disk_loaded),
                      ('Disk error: ', maybe(first_line, self.disk_error))]

        return Group(*(Text.assemble(label, (str(val), 'bold')) for label, val in fields))
    # fmt: on


class TUIConfig(BaseModel):
    using_server: bool = False
    disk_source: str = '/'
    server_source: str = 'localhost:8000'


class InfoScreen(Screen):
    def __init__(self, msg, screen_type='error', name=None, id=None, classes=None):
        super().__init__(name, id, classes)
        self._msg = msg
        self._type = screen_type

    def compose(self):
        colour, label = _attribs[self._type]

        yield MyHeader()
        yield Static(Text.assemble((self._msg, colour)))
        yield Button('Ok')
        yield Footer()

    def on_button_pressed(self, event):
        self.app.pop_screen()


class Border(Widget):
    # foreground-muted?
    DEFAULT_CSS = """Border {
        border: round $foreground;
        border-title-align: left;
        height: auto;
        overflow-x: auto;
        }"""

    def __init__(self, title, *args, **kwargs):
        Widget.__init__(self, *args, **kwargs)
        if title:
            self.border_title = title


class HeaderStatus(Static):
    DEFAULT_CSS = 'HeaderStatus { height: 100%; width: auto; dock: right; }'
    status_summary = reactive('', layout=True)

    def render(self):
        return Text.from_markup(self.status_summary)


class MyHeader(Widget):
    DEFAULT_CSS = """
    MyHeader {
        dock: top;
        width: 100%;
        background: black;
        color: $foreground;
        height: 1;
        margin-bottom: 1;
    }
    """

    mmodel = reactive(None)
    status = reactive(None)
    status_summary = reactive('   ')

    def __init__(self):
        Widget.__init__(self)
        bind(self, 'mmodel', 'status')

    def watch_status(self):
        string = ''
        if self.status:
            string += f'[bright_red]{self.status.error_status()}[/bright_red]'

        task_creator = ModelTaskCreator()

        if self.mmodel:
            from imandra.u.agents.code_logician.base import FormalizationStatus as FS

            # fmt: off
            counts = self.mmodel.summary()['frm_statuses']
            bad = counts[FS.INADMISSIBLE]
            ok = (counts[FS.TRANSPARENT]
                  + counts[FS.EXECUTABLE_WITH_APPROXIMATION]
                  + counts[FS.ADMITTED_WITH_OPAQUENESS])

            def is_in_progress(m): return 1 if task_creator.gen_model_status(m).outstanding_task_ID else 0

            in_progress = sum(map(is_in_progress, self.mmodel.models.values()))
            in_progress_str = f'[blink]{in_progress}[/blink]' if in_progress > 0 else '0'
            string += (f' processing: {in_progress_str} | '
                       f'[dim]? {counts[FS.UNKNOWN]}[/dim] '
                       f'[light_green]✓ {ok}[/light_green] '
                       f'[bright_red]✘ {bad}[/bright_red]')
            # fmt: on

        self.status_summary = string

    def compose(self):
        icon = Text.assemble((' OIO', 'bold #3363FF'), (' IMANDRA', 'bold #00C6CF'))
        icon_widget = HeaderIcon()
        icon_widget.icon = icon  # pyright: ignore[reportAttributeAccessIssue]
        icon_widget.styles.width = 'auto'
        icon_widget.styles.dock = 'left'
        yield icon_widget
        yield HeaderTitle()
        yield HeaderStatus().data_bind(MyHeader.status_summary)  # pyright: ignore

    def on_mount(self):
        def set_title(title):
            try:
                self.query_one(HeaderTitle).update(self.screen.title or self.app.title)
            except Exception as e:
                print(f'???? {e}')

        self.watch(self.screen, 'title', set_title)


def opaques_rich(opaques, limit=None):
    from rich.table import Table

    table = Table(show_header=False, box=None, padding=(0, 1))
    for i, opa in enumerate(opaques[:limit], 1):
        num_assumptions = len(opa.assumptions) if hasattr(opa, 'assumptions') else 0
        has_approx = hasattr(opa, 'approximation') and opa.approximation is not None
        status_icon = (
            '[bright_green]✓[/bright_green]'
            if has_approx
            else '[bright_yellow]○[/bright_yellow]'
        )
        table.add_row(
            f'{i}.',
            f'{status_icon} {opa.name}',
            f'({num_assumptions} assumptions)',
        )
    if limit is not None and len(opaques) > limit:
        table.add_row('...', f'[dim]({len(opaques) - limit} more)[/dim]', '')

    return f'Opaque Functions ({len(opaques)}):', table


class HScroll(ScrollableContainer):
    DEFAULT_CSS = """
    HScroll {
        width: 150;
        height: auto;
        layout: vertical;
        overflow-y: hidden;
        overflow-x: scroll;
    }
    """


def named_vals(name_val_pairs):
    def hg(name, val):
        l = Static(f'[$primary]{name}[/]: ')
        v = Static(str(val))
        v.styles.width = '1fr'
        l.styles.width = 'auto'
        return HorizontalGroup(l, v)

    return VerticalGroup(*(hg(name, val) for name, val in name_val_pairs))


def decomp_ui(decomp):
    from imandra.u.agents.code_logician.base.region_decomp import (
        render_decomp_res_content,
    )
    from rich.panel import Panel

    parts = [  # list[(Optional(x), str, (x -> rich))]
        (
            decomp.raw,
            'Raw Request',
            lambda raw: Group(
                Text(f'Src func name: {raw.src_func_name}'),
                Text(f'IML func name: {raw.iml_func_name}'),
                Text(f'Description: {raw.description}'),
            ),
        ),
        (decomp.data, 'Request Data', lambda d: d.__rich__()),
        (decomp.res, 'Result', lambda c: Group(*render_decomp_res_content(c))),
        (decomp.test_cases, 'Test Cases', decomp.render_test_cases),
    ]
    return Static(Group(*[Panel(f(x), title=t) for x, t, f in parts if x]))


def vg_ui(vg):
    from rich.pretty import Pretty

    def req_repr(req):
        return named_vals(
            [
                ('Src func names', req.src_func_names),
                ('IML func names', req.iml_func_names),
                ('Description', req.description),
                ('Logical statement', req.logical_statement),
            ]
        )

    def data_repr(data):
        return named_vals([('Predicate', data.predicate), ('Kind', data.kind)])

    def maybe_(f, x):
        return maybe_else(Static('None'), f, x)

    def text_repr(x):
        return Text(repr(x))

    return VerticalGroup(
        Border('RawVerifyReq', maybe_(req_repr, vg.raw)),
        Border('VerifyReqData', maybe_(data_repr, vg.data)),
        Border(
            'Result', HScroll(Static(Pretty(vg.res, overflow='ellipsis', no_wrap=True)))
        ),
    )


def text_read(f):
    with open(f) as s:
        return s.read()


def local_file(name):
    return Path(__file__).parent / name



#
#   Imandra Inc.
#
#   codelogician/modeling/model_task_creator.py
#

from __future__ import annotations

import datetime
import logging
import re
from dataclasses import dataclass

from .dep_context import ModelDepContext
from .model import Model
from .status import ModelStatus
from .task import (
    DecompVGsModelTask,
    FormalizationNeed,
    ModelTask,
    TaskKind,
    UserManualIMLEditTask,
)

logger = logging.getLogger(__name__)


def _remove_import_directives(text: str) -> str:
    return re.sub(r'^\s*\[@@@import.*?\]\s*\n?', '', text, flags=re.MULTILINE)


@dataclass(frozen=True)
class ModelTaskCreator:
    user_wait_time: int | None = 5
    src_wait_time: int | None = 5

    # -------------------------
    # Checks / reasons
    # -------------------------

    def user_iml_change_ready(self, model: Model, wait: int | None = None) -> bool:
        if model.user_iml_edit is None:
            logger.debug('[%s] No user IML edit present', model.rel_path)
            return False

        wait_time = self.user_wait_time if wait is None else wait
        if wait_time is None:
            logger.debug(
                '[%s] User IML edit ready immediately (no wait)', model.rel_path
            )
            return True

        delta = datetime.datetime.now() - model.user_iml_edit.time_of_edit
        ready = delta > datetime.timedelta(seconds=wait_time)

        logger.debug(
            '[%s] User IML edit age=%ss wait=%ss ready=%s',
            model.rel_path,
            delta.total_seconds(),
            wait_time,
            ready,
        )
        return ready

    def has_src_code_changed(self, model: Model, wait: int | None = None) -> bool:
        if model.agent_state is None:
            logger.debug('[%s] No agent state → src considered changed', model.rel_path)
            return True

        wait_time = self.src_wait_time if wait is None else wait
        if wait_time is None or model.src_code_last_changed is None:
            long_enough = True
            delta_s: float | None = None
        else:
            delta = datetime.datetime.now() - model.src_code_last_changed
            delta_s = delta.total_seconds()
            long_enough = delta > datetime.timedelta(seconds=wait_time)

        changed = hash(model.src_code) != hash(model.agent_state.src_code)

        logger.debug(
            '[%s] Src change check: last_changed_age_s=%s wait=%s long_enough=%s changed=%s',
            model.rel_path,
            delta_s,
            wait_time,
            long_enough,
            changed,
        )
        return long_enough and changed

    def context_provided(self, model: Model) -> bool:
        provided = bool(model.context)
        logger.debug('[%s] Context provided=%s', model.rel_path, provided)
        return provided

    def formalization_reasons(
        self, model: Model, wait: int | None = None
    ) -> list[FormalizationNeed]:
        if model.src_code is None:
            logger.debug('[%s] No src code → no formalization reasons', model.rel_path)
            return []

        if model.agent_state is None:
            logger.debug('[%s] No agent state → needs formalization', model.rel_path)
            return [FormalizationNeed.NO_AGENT_STATE]

        reasons: list[FormalizationNeed] = []
        if self.has_src_code_changed(model, wait):
            reasons.append(FormalizationNeed.SRC_CODE_CHANGED)
        if self.context_provided(model):
            reasons.append(FormalizationNeed.CONTEXT_ADDED)
        if ModelDepContext.deps_changed(model):
            reasons.append(FormalizationNeed.DEPS_CHANGED)

        logger.debug('[%s] Formalization reasons=%s', model.rel_path, reasons)
        return reasons

    def deps_need_formalization(self, m: Model) -> bool:
        # fmt: off
        def p(d, t):
            r = self.formalization_reasons(d, t)
            if r:
                logger.debug('[%s] Dep %s needs formalization: %s', m.rel_path, d.rel_path, r)
            return r or self.deps_need_formalization(d)

        need = any(p(d, self.user_wait_time) for d in ModelDepContext.deps(m))
        # fmt: on
        logger.debug('[%s] deps_need_formalization=%s', m.rel_path, need)
        return need

    # -------------------------
    # Analysis (VGs / decomps)
    # -------------------------

    def _formalization_complete(self, model: Model) -> bool:
        """
        Conservative check: we consider the model "formalized enough to analyze"
        if we have an agent_state and some IML model exists.
        """
        complete = (model.agent_state is not None) and (model.iml_model() is not None)
        logger.debug('[%s] Formalization complete=%s', model.rel_path, complete)
        return complete

    def _missing_artifacts(self, model: Model) -> tuple[bool, bool]:
        """
        Returns: (need_vgs, need_decomps)

        Policy:
        - Empty results ([], None) can be a valid outcome of ANALYZE.
        - We only consider artifacts "missing" if we have not attempted generation
        since the current formalization baseline.
        """
        if model.agent_state is None:
            logger.debug(
                '[%s] Missing artifacts check skipped: no agent_state', model.rel_path
            )
            return (False, False)

        vgs = getattr(model.agent_state, 'vgs', None)
        decomps = getattr(model.agent_state, 'region_decomps', None)

        need_vgs = not model.analysis_vgs_attempted
        need_decomps = not model.analysis_decomps_attempted

        logger.debug(
            '[%s] Missing artifacts: need_vgs=%s need_decomps=%s '
            '(attempted_vgs=%s attempted_decomps=%s vgs_present=%s decomps_present=%s)',
            model.rel_path,
            need_vgs,
            need_decomps,
            model.analysis_vgs_attempted,
            model.analysis_decomps_attempted,
            bool(vgs),
            bool(decomps),
        )
        return (need_vgs, need_decomps)

    def _analysis_needed(self, model: Model) -> bool:
        """
        Analyze if formalization is complete, deps are stable, and artifacts are missing.
        """
        if not self._formalization_complete(model):
            logger.debug(
                '[%s] Analysis skipped: formalization incomplete', model.rel_path
            )
            return False

        # Don’t analyze if we need (re)formalization or deps are unstable.
        if self.formalization_reasons(model, self.user_wait_time):
            logger.debug('[%s] Analysis skipped: formalization pending', model.rel_path)
            return False
        if self.deps_need_formalization(model):
            logger.debug(
                '[%s] Analysis skipped: deps need formalization', model.rel_path
            )
            return False
        if ModelDepContext.deps_changed(model):
            logger.debug('[%s] Analysis skipped: deps changed', model.rel_path)
            return False

        need_vgs, need_decomps = self._missing_artifacts(model)
        needed = need_vgs or need_decomps
        logger.debug('[%s] Analysis needed=%s', model.rel_path, needed)
        return needed

    # -------------------------
    # Decide + create a single task
    # -------------------------

    def decide_task_kind(self, model: Model) -> TaskKind | None:
        """
        Decide the single highest-priority task for this model.

        Latest policy:

        - USER_IML_UPDATE can preempt an outstanding task if ready.

        - FORMALIZE can preempt an outstanding task if deps are stable AND there is
        a *real* new-change reason (e.g., SRC_CODE_CHANGED / CONTEXT_ADDED / DEPS_CHANGED),
        NOT merely "NO_AGENT_STATE" (which just means we're still waiting for the
        outstanding formalization result).

        - ANALYZE never preempts: only when there is no outstanding task AND artifacts are missing.
        """

        # 1) User IML override (allowed even if outstanding exists)
        if self.user_iml_change_ready(model):
            logger.debug(
                '[%s] Task decision=USER_IML_UPDATE (preempt=%s)',
                model.rel_path,
                model.outstanding_task_ID is not None,
            )
            return TaskKind.USER_IML_UPDATE

        # 2) Compute formalization needs
        reasons = self.formalization_reasons(model, self.user_wait_time)
        needs_formalization = bool(reasons)
        deps_stable = not self.deps_need_formalization(model)

        logger.debug(
            '[%s] Task decision precheck: outstanding=%s needs_formalization=%s deps_stable=%s reasons=%s',
            model.rel_path,
            model.outstanding_task_ID,
            needs_formalization,
            deps_stable,
            reasons,
        )

        # 3) If anything is outstanding:
        #    - allow FORMALIZE only if there's a real-change reason (not NO_AGENT_STATE)
        #    - never ANALYZE
        if model.outstanding_task_ID is not None:
            # If we're already running a FORMALIZE task, do not spam another one.
            if model.outstanding_task_kind is TaskKind.FORMALIZE:
                return None

            if needs_formalization and deps_stable:
                # "NO_AGENT_STATE" is not a preemption trigger: it just means we're still waiting
                # for the already-outstanding formalization task to produce an agent_state.
                real_change_reasons = [
                    r for r in reasons if r is not FormalizationNeed.NO_AGENT_STATE
                ]

                if real_change_reasons:
                    logger.debug(
                        '[%s] Task decision=FORMALIZE (preempt=True real_change_reasons=%s all_reasons=%s)',
                        model.rel_path,
                        real_change_reasons,
                        reasons,
                    )
                    return TaskKind.FORMALIZE

            logger.debug(
                '[%s] Task decision=NONE (blocked by outstanding_task_ID=%s; reasons=%s)',
                model.rel_path,
                model.outstanding_task_ID,
                reasons,
            )
            return None

        # 4) Normal formalize when idle
        if needs_formalization and deps_stable:
            logger.debug(
                '[%s] Task decision=FORMALIZE (preempt=False reasons=%s)',
                model.rel_path,
                reasons,
            )
            return TaskKind.FORMALIZE

        # 5) Analyze only when idle
        if self._analysis_needed(model):
            logger.debug('[%s] Task decision=ANALYZE', model.rel_path)
            return TaskKind.ANALYZE

        logger.debug('[%s] Task decision=NONE', model.rel_path)
        return None

    def create_task(self, model: Model) -> ModelTask | UserManualIMLEditTask | None:
        """
        Create at most one task.
        """

        if model.src_code is None:
            logger.debug(
                '[%s] Skipping task creation: src_code is None (deleted)',
                model.rel_path,
            )
            return None

        k = self.decide_task_kind(model)
        if k is None:
            logger.debug('[%s] No task created', model.rel_path)
            return None

        logger.debug('[%s] Creating task kind=%s', model.rel_path, k)

        if k is TaskKind.USER_IML_UPDATE:
            return self._user_iml_update_task(model)

        if k is TaskKind.FORMALIZE:
            return self._formalization_task(model)

        if k is TaskKind.ANALYZE:
            return self._analysis_task(model)

        raise RuntimeError(f'Unhandled task kind: {k}')

    # -------------------------
    # Task constructors
    # -------------------------

    def _formalization_task(self, model: Model) -> ModelTask:
        context = """Do not use `[@@measure ...]` annotations on nested functions (closures). If a function needs a `[@@measure ...]` annotation, structure the code so that the nested function is moved to the top-level.

When writing a recursive function where the recursion is controlled by an integer argument, put that integer argument as the first argument. Also try to arrange for the integer argument to count down to zero.

Use the operator `=` to test equality between numbers of type `Real`, not the operator `=.`, which does not exist."""

        task = ModelTask(
            rel_path=model.rel_path,
            src_code=model.src_code or '',
            context=f'{context}\n{model.context}',
            graph_state=model.agent_state.graph_state if model.agent_state else None,
            dependencies=ModelDepContext.dependencies_iml_models(model),
        )
        model.outstanding_task_ID = task.task_id
        logger.debug('[%s] Created FORMALIZE task_id=%s', model.rel_path, task.task_id)
        return task

    def _user_iml_update_task(self, model: Model) -> UserManualIMLEditTask | None:
        if model.user_iml_edit is None:
            logger.debug(
                '[%s] USER_IML_UPDATE requested but no user_iml_edit', model.rel_path
            )
            return None

        task = UserManualIMLEditTask(
            rel_path=model.rel_path,
            graph_state=model.agent_state.graph_state if model.agent_state else None,
            iml_code=_remove_import_directives(model.user_iml_edit.user_iml_entry),
            dependencies=ModelDepContext.dependencies_iml_models(model),
        )
        model.outstanding_task_ID = task.task_id
        logger.debug(
            '[%s] Created USER_IML_UPDATE task_id=%s', model.rel_path, task.task_id
        )
        return task

    def _analysis_task(self, model: Model) -> DecompVGsModelTask:
        need_vgs, need_decomps = self._missing_artifacts(model)

        task = DecompVGsModelTask(
            rel_path=model.rel_path,
            src_code=model.src_code or '',
            context=model.context,
            graph_state=model.agent_state.graph_state if model.agent_state else None,
            dependencies=ModelDepContext.dependencies_iml_models(model),
            gen_vgs=need_vgs,
            gen_decomps=need_decomps,
        )
        model.outstanding_task_ID = task.task_id
        logger.debug(
            '[%s] Created ANALYZE task_id=%s gen_vgs=%s gen_decomps=%s',
            model.rel_path,
            task.task_id,
            need_vgs,
            need_decomps,
        )
        return task

    # -------------------------
    # Status
    # -------------------------

    def gen_model_status(self, m: Model) -> ModelStatus:
        status = ModelStatus(
            src_code_status=m.src_code_status(),
            instructions_added=m.context_provided(),
            deps_changed=ModelDepContext.deps_changed(m),
            deps_need_formalization=self.deps_need_formalization(m),
            outstanding_task_ID=m.outstanding_task_ID,
            outstanding_task_kind=m.outstanding_task_kind,
            formalization_status=m.formalization_status(),
        )
        logger.debug('[%s] ModelStatus=%s', m.rel_path, status)
        return status
